import h5py
from pathlib import Path
import re
import sys
import warnings

import numpy as np 

from npyx.utils import assert_int, assert_float
from npyx.inout import get_npix_sync, chan_map, extract_rawChunk, read_metadata
from npyx.spk_t import ids, trn, trn_filtered
from npyx.spk_wvf import wvf_dsmatch
from npyx.gl import get_units, check_periods

def print_h5_contents(h5_path, txt_output=False):
    """
    h5_path: str, path to .h5 file
    txt_output: bool, if True prints contents to file
                      (same name as h5 name_content.txt)
    """
    h5_path = Path(h5_path)
    if txt_output:
        txt_output_path = h5_path.parent / f"{h5_path.name[:-3]}_content.txt"
    with h5py.File(h5_path, "a") as hdf:
        if txt_output:
            with open(txt_output_path, "w") as txt:
                original_stdout = sys.stdout
                sys.stdout = txt
                visititems(hdf, visitor_func)
                sys.stdout = original_stdout
        else:
            visititems(hdf, visitor_func)

def label_optotagged_unit_h5(h5_path, dataset, unit, label):
    """
    Add optotagged label to neuron.

    - h5_path: full path to h5 file
    - dataset: str, neuron dataset (yy-mm-dd_mouse_probex)
    - unit: neuron unit index
    - label: label to add
    """
    authorized_labels = ["PkC_ss", "PkC_cs", "MLI", "MFB", "GoC", "GrC"]
    assert label in authorized_labels
    add_data_to_unit_h5(h5_path, dataset, unit, label, 'optotagged_label') 

def add_unit_h5(h5_path, dp, unit, lab_id, periods=[[0,20*60]],
                unit_abolute_id=None, sync_chan_id=None,
                again=False, again_wvf=False, plot_debug=False, verbose=False,
                dataset=None, snr_window=[0.1, 30.1], raw_snippet_halfrange=2,
                optostims=None, optostims_threshold=None,
                **kwargs):
    """
    Add a Kilosort sorted unit to an HDF5 file.

    Adds a Kilosort unit to a new or existing HDF5 five file using the
    file format specified by the C4 collaboration.

    Each unit can be accessed from 2 paths which point to the same data:
    - an absolute path, {unit_abolute_id}/
                        which allows for a flat hierarchy and makes it easier to compute features,
                        easier to work on data from other labs
    - a relative path, datasets/{dataset}/{unit}/
                       which allows to index units per dataset,
                       easier to work on your own data

    Example:
        add_unit_h5('my_lab_data.h5', '/path/to/kilosort_results', 1, lab_id='pi_last_name')
    Adds the unit with id 1 to the HDF5 file in the current directory called 'my_lab_data.h5'.
    Additional units can be added by calling the same function with identical arguments,
    but incrementing the unit id field.

    Required parameters:
    - h5_path: Path to the h5 file to create/append
    - dp: Path the Kilosort data directory
    - unit: The unit id/neuron unit index
    - lab_id: The lab/PI id to use to label the units
    - periods: 'all' or [[t1,t2],[t3,t4],...] in seconds

    Key-value parameters:
    - unit_absolute_id: unit absolute id. Will increment from the last unit added to h5 file.
    - sync_chan_id: The channel id used to denote opto stimulation. Defaults to None.
    - again: Whether to use cached results for storage in the HDF5 file (defaults to False)
    - again_wvf: Whether to recompute drift-shift matched waveform in particular (very computationally intensive, defaults to False)
    - verbose: Additional verbosity/progress
    - dataset: A unique ID for this dataset. By default this value is None, in which case
      the dataset id is assumed to the dirname of the data directory passed as the dp argument
    - snr_window: A two item list containing the start and stop times (in seconds) for computation of the
      snr/voltage clip/waveform results
    - raw_snippet_halfrange: int, range of channels around peak channel to consider for the snippet of raw data (max 10)
    - optostims: an optional 2D array (n_stims, 2) containing the optostimuli times in seconds
                 (1st column: onsets, 2nd column: offsets). By default None, will be read from sync channel (at sync_chan_id)
    - optostims_threshold: float, time before which optostims will be ignored
                           (handles sync signals without light at beginning of recording)

    Additional key-value parameteters:
    - *any_key* = *any_value*
    All additional key-value parameters passed to this function are stored in the HDF5 file.
    Therefore, custom keys can be stored in the HDF5 file should additional
    information be required for an individual neuron. E.g., calling
        add_unit_h5('my_lab_data.h5', '/path_to_kilosort_results', 1, my_note="Cool info")
    will result in a key of 'my_note' and a value of "Cool info" being stored in the HDF5 file
    for this unit.
    """
    dp=Path(dp)
    meta = read_metadata(dp) 
    samp_rate = meta['highpass']['sampling_rate']

    # hard-coded parameters
    waveform_samples = 6  # ms
    waveform_samples = int(waveform_samples*samp_rate/1000)

    # open file in append mode
    h5_file = h5py.File(h5_path, "a")

    # check whether neuron already exists in dataset
    if dataset is None:
        dataset = dp.name
    check_dataset_format(dataset)
    neuron_path = f"datasets/{dataset}/{unit}"
    if neuron_path in h5_file:
        neuron_absolute_path=h5_file[f'{neuron_path}/neuron_absolute_id'][()].decode()
        if again:
            del h5_file[neuron_path]
            del h5_file[neuron_absolute_path]
        else:
            print(f"Neuron found in h5 file: {neuron_path} ({neuron_absolute_path})")
            neuron_group = h5_file[neuron_absolute_path]
            #return neuron_path
        
    if neuron_path not in h5_file or again:
        # figure out where we're at
        if unit_abolute_id is None:
            if f"{lab_id}_neuron_0" not in h5_file:
                unit_abolute_id = 0
            else:
                root_groups = list(h5_file.keys())
                neuron_ids = [int(x.split('_')[-1]) for x in root_groups if f"{lab_id}_neuron" in x]
                unit_abolute_id = np.sort(neuron_ids)[-1] + 1

        # create group for new neuron
        neuron_absolute_path=f'{lab_id}_neuron_{unit_abolute_id}'
        neuron_group = h5_file.create_group(neuron_absolute_path)
        h5_file[neuron_path] = neuron_group
        print(f"Adding data at {neuron_path} ({neuron_absolute_path})...")

    # metadata
    add_dataset_to_group(neuron_group, 'lab_id', lab_id, again)
    add_dataset_to_group(neuron_group, 'dataset_id', dataset, again)
    add_dataset_to_group(neuron_group, 'neuron_id', unit, again)
    add_dataset_to_group(neuron_group, 'neuron_absolute_id', neuron_group.name, again)
    add_dataset_to_group(neuron_group, 'sampling_rate', samp_rate, again)
    
    # add any additional keys passed to this function
    for key, value in kwargs.items():
        add_dataset_to_group(neuron_group, key, value, again)

    # spike_times
    periods = check_periods(periods)
    if 'spike_indices' not in neuron_group or again:
        t = trn(dp, unit, periods=periods, again=again)
        add_dataset_to_group(neuron_group, 'spike_indices', t, again)
        if optostims is None:
            ons, offs = get_npix_sync(dp, verbose=False)
            if sync_chan_id is None:
                sync_chan_id = get_stim_chan(ons)
            ons, offs = ons[sync_chan_id], offs[sync_chan_id]
            if ons[0] > offs[0]:
                ons, offs = offs, ons 
            if len(offs) == len(ons) - 1:
                offs = np.append(offs, meta['recording_length_seconds'])
            optostims = np.hstack([ons[:, None], offs[:, None]])
        if optostims_threshold is not None:
            opto_m = optostims[:,0] > optostims_threshold
            optostims = optostims[opto_m,:]
        add_dataset_to_group(neuron_group, 'optostims', optostims, again)
        # Only consider spikes 10s before first opto onset
        sane_spikes = (t < (optostims[0,0]-10)*samp_rate)
        add_dataset_to_group(neuron_group, 'sane_spikes', sane_spikes, again)
        
    if 'fn_fp_filtered_spikes' not in neuron_group or again:
        # get good spikes mask for all spikes
        # because trn_filtered can only work on a contiguous chunk
        if periods is 'all':
            periods_m_range = [0, meta['recording_length_seconds']/60]
        else:
            periods_m_range = [periods.min()/60, periods.max()/60]
        fp_fn_good_spikes = trn_filtered(dp, unit, plot_debug=plot_debug, again=again, period_m=periods_m_range)[1]


        # if periods is not all, trim down the mask to spikes in periods
        if periods is not 'all':
            t = trn(dp, unit, periods=periods) # if again, as recomputed just above anyway, so don't pass the argument
            t_all = trn(dp, unit) # grab all spikes
            periods_mask = np.isin(t_all, t)
            fp_fn_good_spikes = fp_fn_good_spikes[periods_mask]

        add_dataset_to_group(neuron_group, 'fn_fp_filtered_spikes', fp_fn_good_spikes, again)

    # waveforms
    if 'mean_waveform_preprocessed' not in neuron_group\
        or ('amplitudes' not in neuron_group)\
        or ('voltage_sample' not in neuron_group)\
        or again: # must recompute chan_bottom and chan_top - suboptimal, can be rewritten
        dsm_tuple = wvf_dsmatch(dp, unit, t_waveforms=waveform_samples, periods=periods,
                                again=again_wvf, plot_debug=plot_debug, verbose=verbose, n_waves_used_for_matching=500)
        dsm_waveform, peak_chan = dsm_tuple[1], dsm_tuple[3]
        add_dataset_to_group(neuron_group, 'primary_channel', peak_chan)
        chan_bottom = max(0, peak_chan-11)
        chan_top = min(383, peak_chan+11)
        dsm_waveform_chunk = dsm_waveform[:, chan_bottom:chan_top]
        add_dataset_to_group(neuron_group, 'mean_waveform_preprocessed', dsm_waveform_chunk.T, again)
        add_dataset_to_group(neuron_group, 'consensus_waveform', dsm_waveform_chunk.T*np.nan, again)
        cm = chan_map(dp)
        add_dataset_to_group(neuron_group, 'channel_ids', np.arange(chan_bottom, chan_top, dtype=np.dtype('uint16')), again)
        add_dataset_to_group(neuron_group, 'channelmap', cm[chan_bottom:chan_top, 1:2], again)

    chunk = None
    if ('amplitudes' not in neuron_group) or ('voltage_sample' not in neuron_group) or again:
        chunk = extract_rawChunk(dp, snr_window, channels=np.arange(chan_bottom, chan_top), 
                                 scale=False, whiten=False, hpfilt=False, verbose=False)

    # quality metrics
    if 'amplitudes' not in neuron_group or again:
        add_dataset_to_group(neuron_group, 'amplitudes', np.load(dp/'amplitudes.npy').squeeze()[ids(dp, unit, periods=periods)], again)
        mad = np.median(np.abs(chunk) - np.median(chunk, axis=1)[:, None], axis=1) 
        std_estimate = (mad / 0.6745) # Convert to std
        add_dataset_to_group(neuron_group, 'channel_noise_std', std_estimate, again)
    
    # voltage snippets
    if 'voltage_sample' not in neuron_group or again:
        # Only store the voltage sample for the primary channel
        peak_chan = neuron_group['primary_channel']
        raw_snippet_halfrange = np.clip(raw_snippet_halfrange, 0, 10)
        c1, c2 = max(0,int(chunk.shape[0]/2-raw_snippet_halfrange)), min(chunk.shape[0]-1, int(chunk.shape[0]/2+raw_snippet_halfrange+1))
        raw_snippet = chunk[c1:c2,:]
        add_dataset_to_group(neuron_group, 'voltage_sample', raw_snippet) # still centered on peak channel, but half the size
        add_dataset_to_group(neuron_group, 'voltage_sample_start_index', int(snr_window[0] * samp_rate))
        add_dataset_to_group(neuron_group, 'scaling_factor', meta['bit_uV_conv_factor']) 

    # layer
    add_dataset_to_group(neuron_group, 'phyllum_layer', 0, again)
    add_dataset_to_group(neuron_group, 'human_layer', 0, again)

    # ground truth labels
    add_dataset_to_group(neuron_group, 'expert_label', 0, again)
    add_dataset_to_group(neuron_group, 'optotagged_label', 0, again)

    # predicted labels
    add_dataset_to_group(neuron_group, 'lisberger_label', 0, again)
    add_dataset_to_group(neuron_group, 'hausser_label', 0, again)
    add_dataset_to_group(neuron_group, 'medina_label', 0, again)

    # close file
    h5_file.close()

    return neuron_path

def add_data_to_unit_h5(h5_path, dataset, unit, data, field):
    """
    Add data to neuron already in h5 file.

    - h5_path: full path to h5 file
    - dataset: str, neuron dataset (yy-mm-dd_mouse_probex)
    - unit: unit index
    - data: data to add to unit
    - field: name of dataset to add data (id exists already, will overwrite)
    """
    with h5py.File(h5_path, "a") as h5_file:
        check_dataset_format(dataset)
        neuron_path = f"datasets/{dataset}/{unit}"
        assert neuron_path in h5_file, f"WARNING unit {neuron_path} does not seem to be present in the file. To add it, use add_unit_h5()."
        add_dataset_to_group(h5_file[neuron_path], field, data, True)

def add_dataset_to_group(group, dataset, data, again=False):
    if dataset in group:
        if again:
            del group[dataset]
        else:
            return
    group[dataset] = data
    return
    
def add_units_to_h5(h5_path, dp, **kwargs):
    """
    Add all units at the respective data path to an HDF5 file.

    This is a high-level function designed to add all units at the
    specified datapath to an HDF5 file. All additional key-value 
    arguments are passed to `add_unit_h5`

    Example:
      add_units_to_h5('my_lab_data.h5', '/path/to/kilosort_results', lab_id='pi_last_name')
    Will add all sorted units in the 'kilosort_results' directory 
    to the HDF5 file called 'my_lab_data.h5' (in the current directory).
    """
    for unit_id in get_units(dp):
        add_unit_h5(h5_path, dp, unit_id, **kwargs)

def get_stim_chan(ons, min_th=20):
    chan = -1
    for k, v in ons.items():
        if len(v) > min_th:
            chan = k
    assert chan != -1
    return chan

def visititems(group, func):
    with h5py._hl.base.phil:
        def proxy(name):
            """ Call the function with the text name, not bytes """
            name = group._d(name)
            return func(name, group[name])
        return group.id.links.visit(proxy)

def visitor_func(name, node):
    if isinstance(node, h5py.Dataset):
        n=node[()]
        if isinstance(n, bytes):
            s=n.decode()
        elif isinstance(n, np.ndarray):
            s=f"ndarray {n.shape}"
        elif assert_int(n) or assert_float(n):
            s=n
        else:
            s=type(n)
        string = f"{name}: {s}"
    else:
        string = name
    print(string)

def check_dataset_format(dataset):
    """
    Checks whether dataset name is formatted properly
    i.e. aa-mm-dd_iiXXX_probeX (mouse can only be names )
    """
    warning = "WARNING last folder of path should match format: aa-mm-dd_ii[0-1000]_probe[0-9] (ii = 2 [a-z] initials)"
    pattern = "[0-9]{2}-[0-9]{2}-[0-9]{2}_[a-z]{2}[0-9]{3}_probe[0-9]"
    if re.match(pattern, dataset, re.IGNORECASE) is None:
        warnings.warn(warning)


# -*- coding: utf-8 -*-

import os
from pathlib import Path
from tracemalloc import start

from numba import njit
from numba.typed import List
from numba.core.errors import NumbaDeprecationWarning, NumbaPendingDeprecationWarning
import warnings

warnings.simplefilter("default", category=NumbaDeprecationWarning) #'ignore'
warnings.simplefilter('default', category=NumbaPendingDeprecationWarning)#'ignore'

from ast import literal_eval as ale
import numpy as np
from numpy.fft import rfft, irfft
import matplotlib.pyplot as plt

from six import integer_types
from statsmodels.nonparametric.smoothers_lowess import lowess
import scipy.stats as stt
import scipy.signal as sgnl

import logging
from math import pi, log
from scipy import fft, ifft
from scipy.optimize import curve_fit
from scipy.signal import cspline1d_eval, cspline1d

#%% Colors dictionnaries

phyColorsDic = {
    0:(53./255, 127./255, 255./255),
    1:(255./255, 0./255, 0./255),
    2:(255./255,215./255,0./255),
    3:(238./255, 53./255, 255./255),
    4:(84./255, 255./255, 28./255),
    5:(255./255,165./255,0./255),
    -1:(0., 0., 0.),
    }

mpl_colors=plt.rcParams['axes.prop_cycle'].by_key()['color']

DistinctColors20 = [(127,127,127),(0,0,143),(182,0,0),(0,140,0),(195,79,255),(1,165,202),(236,157,0),(118,255,0),(255,127,0),
    (255,117,152),(148,0,115),(0,243,204),(72,83,255),(0,127,255),(0,67,1),(237,183,255),(138,104,0),(97,0,163),(92,0,17),(255,245,133)]
DistinctColors20 = [(c[0]/255, c[1]/255, c[2]/255) for c in DistinctColors20]
DistinctColors15 = [(127,127,127),(255,255,0),(0,0,143),(255,0,0),(50,255,255),(255,0,255),(94,0,33),(0,67,0),
    (255,218,248),(0,178,0),(124,72,255),(211,145,0),(5,171,253),(126,73,0),(147,0,153)]
DistinctColors15 = [(c[0]/255, c[1]/255, c[2]/255) for c in DistinctColors15]

mark_dict = {
".":"point",
",":"pixel",
"o":"circle",
"v":"triangle_down",
"^":"triangle_up",
"<":"triangle_left",
">":"triangle_right",
"1":"tri_down",
"2":"tri_up",
"3":"tri_left",
"4":"tri_right",
"8":"octagon",
"s":"square",
"p":"pentagon",
"*":"star",
"h":"hexagon1",
"H":"hexagon2",
"+":"plus",
"D":"diamond",
"d":"thin_diamond",
"|":"vline",
"_":"hline"
}


#%% Utils

_ACCEPTED_ARRAY_DTYPES = (np.float, np.float32, np.float64,
                          np.int, np.int8, np.int16, np.uint8, np.uint16,
                          np.int32, np.int64, np.uint32, np.uint64,
                          np.bool)

def assert_float(x):
    return isinstance(x, (float, np.float,
                          np.float16, np.float32, np.float64))

def assert_int(x):
    return isinstance(x, (int, np.int, np.int8, np.int16,
                          np.uint8, np.uint16, np.int32,
                          np.int64, np.uint32, np.uint64))

def assert_iterable(x):
    return hasattr(x, '__iter__')

def npa(arr=[], **kwargs):
    '''Returns np.array of some kind.
    Optional params:
        - zeros: tuple. If provided, returns np.zeros(zeros)
        - ones: tuple. If provided, returns np.ones(ones)
        - empty: tuple. If provided, returns np.empty(empty)
        - dtype: numpy datatype. If provided, returns np.array(arr, dtype=dtype) .'''

    dtype=kwargs['dtype'] if 'dtype' in kwargs.keys() else None
    if 'zeros' in kwargs.keys():
        arr = np.zeros(kwargs['zeros'], dtype=dtype)
    elif 'ones' in kwargs.keys():
        arr = np.ones(kwargs['ones'], dtype=dtype)
    elif 'empty' in kwargs.keys():
        arr = np.empty(kwargs['empty'], dtype=dtype)
    else:
        arr=np.array(arr, dtype=dtype)
    return arr

def isnumeric(x):
    x=str(x).replace('−','-')
    try:
        ale(x)
        return True
    except:
        return False

def sign(x):
    "Returns the sign of the input number (1 or -1). 1 for 0 or -0."
    x=npa(x)
    x[x==0]=1
    return (x/np.abs(x)).astype(np.int64)

def minus_is_1(x):
    return abs(1-1*x)*1./2

def read_pyfile(filepath, ignored_chars=[" ", "'", "\"", "\n", "\r"]):
    '''
    Reads .py file and returns contents as dictionnary.

    Assumes that file only has "variable=value" pairs, no fucntions etc

    - filepath: str, path to file
    - ignored_chars: list of characters to remove (only trailing and leading)
    '''
    filepath = Path(filepath)
    assert filepath.exists(), f'{filepath} not found!'

    params={}
    with open(filepath) as f:
        for ln in f.readlines():
            assert '=' in ln, 'WARNING read_pyfile only works for list of variable=value lines!'
            tmp = ln.split('=')
            for i, string in enumerate(tmp):
                string=string.strip("".join(ignored_chars))
                tmp[i]=string
            k, val = tmp[0], tmp[1]
            try: val = ale(val)
            except: pass
            params[k]=val

    return params

def list_files(directory, extension, full_path=False):
    directory=str(directory)
    files = [f for f in os.listdir(directory) if f.endswith('.' + extension)]
    files.sort()
    if full_path:
        return [Path('/'.join([directory,f])) for f in files]
    return files

def any_n_consec(X, n_consec, where=False):
    '''
    The trick below finds whether there are n_consec consecutive ones in the array comp
    by adding each element with its neighbour.
    At least 2 consec ones: add them 1 time and check if there is a 2=2**1 somewhere.
    At least 2 consec ones: add them 3 times and check if there is a 4=2**2 somewhere.
    ...
    Parameters:
        - X: array of booleans or binary
        - n_consec: int, number of consecutive True/False values to find
        - where: bool, returns array of indices if True
    '''
    X=npa(X).astype(np.int)
    assert np.all((X==0)|(X==1)), 'X array should be only 0s and 1s!'
    for i in range(n_consec-1):
        X = X[:-1]+X[1:]
    b=np.any(X==2**(n_consec-1))
    if not where:
        return b
    if b:
        i=np.nonzero(X==2**(n_consec-1))[0]
        i_edges1=i[np.append(0,np.diff(i))!=1]
        i_edges2=(i+n_consec-1)[np.append(np.diff(i), 0)!=1]
        return b, [np.arange(i_edges1[e],i_edges2[e]+1) for e in range(len(i_edges1))]
    return b, npa([])

def thresh_consec0(arr, th, n_consec, sgn=1, exclude_edges=True):
    '''
    SLOWER THAN thresh_consec AND FORCED TO PROVIDE N_CONSEC -> SHITTY
    Returns indices and values of threshold crosses lasting >=n_consec consecutive samples in arr.
    Parameters:
        - arr: numpy array to threshold
        - th: float, threshold
        - n_consec: int, minimum number of consecutive elements beyond threshold
        - sgn: int, positive (for cases above threshold) or negative (below threshold)
        - exclude_edges: bool, if true edges of arr are not considered as threshold crosses
                         in cases where arr starts or ends beyond threshold
    '''
    arr=npa(arr)
    if not arr.ndim==1:
        assert 1 in arr.shape
        arr=arr.flatten()

    arr= (arr-th)*sign(sgn)+th # Flips the array around threshold if sgn==-1
    comp = (arr>=th)

    crosses=any_n_consec(comp, n_consec, where=True)[1]

    if exclude_edges:
        if crosses[0][0]==0: # starts beyond threshold and lasts >=n_consec_bins
            crosses=crosses[1:]
        if crosses[-1][-1]==len(arr)-1: # starts beyond threshold and lasts >=n_consec_bins
            crosses=crosses[:-1]

    return [np.vstack([cross, arr[cross]]) for cross in crosses]

@njit(cache=True)
def thresh_numba(arr, th, sgn=1, pos=1):
    '''Returns indices of the data points closest to a directed crossing of th.
    - data: numpy array.
    - th: threshold value.
    - sgn: 1 or -1, direction of the threshold crossing, either positive (1) or negative (-1).
    - pos: 1 or -1, position of closest value, just following or just preceeding.'''
    assert pos in [-1,1]
    assert sgn in [-1,1]
    arr=np.asarray(arr)
    tp=arr.dtype
    assert arr.ndim==1
    arr= (arr-th)*sgn+th # Flips the array around threshold if sgn==-1

    i=np.nonzero(arr>=th)[0] if pos==1 else np.nonzero(arr<=th)[0]
    # If no value is above the threshold or all of them are already above the threshold
    if len(i)==0 or len(i)==len(arr):
        return np.zeros(0).astype(tp)
    if pos==1:
        clip=np.maximum(0, np.minimum(i-1, len(arr)-1))
        return (i[arr[clip]<th]).astype(tp)
    else:
        clip=np.maximum(0, np.minimum(i+1, len(arr)-1))
        return (i[arr[clip]>th]).astype(tp)

def thresh(arr, th, sgn=1, pos=1):
    '''Returns indices of the data points closest to a directed crossing of th.
    - data: numpy array.
    - th: threshold value.
    - sgn: 1 or -1, direction of the threshold crossing, either positive (1) or negative (-1).
    - pos: 1 or -1, position of closest value, just following or just preceeding.'''
    assert pos in [-1,1]
    assert sgn in [-1,1]
    arr=np.asarray(arr).copy()
    assert arr.ndim==1
    arr = (arr-th)*sgn+th # Flips the array around threshold if sgn==-1

    i=np.nonzero(arr>=th)[0] if pos==1 else np.nonzero(arr<=th)[0]
    # If no value is above the threshold or all of them are already above the threshold
    if len(i)==0 or len(i)==len(arr):
        return np.array([])
    return  i[arr[np.clip(i-1, 0, len(arr)-1)]<th] if pos==1 else i[arr[np.clip(i+1, 0, len(arr)-1)]>th]

def thresh_fast(arr, th, sgn=1, pos=1):
    '''Returns indices of the data points closest to a directed crossing of th.
    - data: numpy array.
    - th: threshold value.
    - sgn: 1 or -1, direction of the threshold crossing, either positive (1) or negative (-1).
    - pos: 1 or -1, position of closest value, just following or just preceeding.'''
    assert pos in [-1,1]
    assert sgn in [-1,1]
    assert arr.ndim==1
    m=(arr>th).astype(np.int8)
    ths=np.nonzero(np.diff(m)==sgn)[0]
    if pos: ths+=1
    return ths

def thresh_consec(arr, th, sgn=1, n_consec=0, exclude_edges=True, only_max=False, ret_values=True):
    '''
    Returns indices and values of threshold crosses lasting >=n_consec consecutive samples in arr.
    Parameters:
        - arr: numpy array to threshold
        - th: float, threshold
        - sgn: 1 or -1, positive (for cases above threshold) or negative (below threshold)
        - n_consec: optional int, minimum number of consecutive elements beyond threshold | Defult 0 (any cross)
        - exclude_edges: bool, if true edges of arr are not considered as threshold crosses
                         in cases where arr starts or ends beyond threshold
        - only_max: bool, if True returns only the most prominent threshold cross.
        - ret_values: bool, whether to return crosses values (list of 2d np arrays [indices, array values] of len Ncrosses)
                            rather than mere crosses indices (2d np array of shape (Ncrosses, 2))
    Returns:
        - crosses values, list of 2d np arrays [indices, array values] if ret_values is True
               or indices, 2d np array of shape (Ncrosses, 2) if ret_values is False
    '''
    def thresh_cons(arr, th, sgn=1, n_consec=0, exclude_edges=True, ret_values=True):
        arr=npa(arr)
        if not arr.ndim==1:
            assert 1 in arr.shape
            arr=arr.flatten()

        assert sgn in [-1,1]
        arr= (arr-th)*sgn+th # Flips the array around threshold if sgn==-1

        cross_thp, cross_thn = thresh(arr, th, 1, 1), thresh(arr, th, -1, -1)
        if exclude_edges:
            if len(cross_thp)+len(cross_thn)<=1: cross_thp, cross_thn = [], [] # Only one cross at the beginning or the end e.g.
            else:
                flag0,flag1=False,False
                if cross_thp[-1]>cross_thn[-1]: flag1=True # if + cross at the end
                if cross_thn[0]<cross_thp[0]: flag0=True # if - cross at the beginning
                if flag1: cross_thp=cross_thp[:-1] # remove last + cross
                if flag0: cross_thn=cross_thn[1:] # remove first - cross
        else:
            if len(cross_thp)+len(cross_thn)<=1: cross_thp, cross_thn = [], [] # Only one cross at the beginning or the end e.g.
            else:
                flag0,flag1=False,False
                if cross_thp[-1]>cross_thn[-1]: flag1=True # if + cross at the end
                if cross_thn[0]<cross_thp[0]: flag0=True # if - cross at the beginning
                if flag1: cross_thn=np.append(cross_thn, [len(arr)-1]) # add fake - cross at the end
                if flag0: cross_thp=np.append([0],cross_thp) # add fake + cross at the beginning

        assert len(cross_thp)==len(cross_thn)

        if ret_values or only_max:
            crosses=[np.vstack([np.arange(cross_thp[i], cross_thn[i]+1, 1), ((arr-th)*sgn+th)[cross_thp[i]:cross_thn[i]+1]]) for i in range(len(cross_thp)) if cross_thn[i]+1-cross_thp[i]>=n_consec]
        else:
            crosses=[[cross_thp[i], cross_thn[i]] for i in range(len(cross_thp)) if cross_thn[i]+1-cross_thp[i]>=n_consec]
        return crosses

    sgn=[-1,1] if sgn==0 else [sgn]
    crosses=[]
    for s in sgn:
        crosses+=thresh_cons(arr, th, s, n_consec, exclude_edges, ret_values)

    if only_max and len(crosses)>0:
        cross=crosses[0]
        for c in crosses[1:]:
            if max(abs(c[1,:]))>max(abs(cross[1,:])): cross = c
        if ret_values:
            crosses=[cross]
        else:
            crosses=[[cross[0,0], cross[0,-1]]]

    return crosses

def zscore(arr, frac=4./5, mn_ext=None, sd_ext=None):
    '''
    Returns z-scored (centered, reduced) array using outer edges of array to compute mean and std.
    Parameters:
        - arr: 1D np array
        - frac: fraction of array used to compute mean and standard deviation
        - mn_ext: optional, provide mean computed outside of function
        - sd_ext: optional, provide standard deviation computed outside of function
    '''
    assert 0<frac<=1, 'Z-score fraction should be between 0 and 1!'
    mn = np.mean(np.append(arr[:int(len(arr)*frac/2)], arr[int(len(arr)*(1-frac/2)):])) if mn_ext is None else mn_ext
    sd = np.std(np.append(arr[:int(len(arr)*frac/2)], arr[int(len(arr)*(1-frac/2)):])) if sd_ext is None else sd_ext
    if sd==0: sd=1
    return (arr-mn)*1./sd

def smooth(arr, method='gaussian_causal', sd=5, axis=1, gamma_a=5):
    '''
    Smoothes a 1D array or a 2D array along specified axis.
    Parameters:
        - arr: ndarray/list, array to smooth
        - method: string, see methods implemented below | Default 'gaussian'
        - sd: int, gaussian window sd (in unit of array samples - e.g. use 10 for a 1ms std if bin size is 0.1ms) | Default 5
        - axis: int axis along which smoothing is performed.
        - a_gamma: sqrt of Gamma function rate (essentially std) | Default 5

    methods implemented:
        - gaussian
        - gaussian_causal
        - gamma (is causal)
    '''
    assert arr.ndim<=2,\
        "WARNING this function runs on 3D arrays but seems to shift data leftwards - not functional yet."
    if arr.ndim==1: axis=0
    
    ## Checks and formatting
    assert method in ['gaussian', 'gaussian_causal', 'gamma']
    assert type(sd) in [int, np.int]


    ## pad array at beginning and end to prevent edge artefacts
    C = arr.shape[axis]//2
    pad_width = [[C,C] if i==axis else [0,0] for i in range(arr.ndim)]
    padarr=np.pad(arr, pad_width, 'symmetric')


    ## Compute the kernel
    if method in ['gaussian', 'gaussian_causal']:
        X=np.arange(-4*sd, 4*sd+1)
        kernel=stt.norm.pdf(X, 0, sd)
        if method=='gaussian_causal':
            kernel[:len(kernel)//2]=0
    elif method=='gamma':
        # a = shape, b = scale = 1/rate. std=sqrt(a)/b = sqrt(a) for b=1
        X=np.arange(gamma_a**2//2, max((gamma_a**2)*3//2+1, 10))
        kernel=stt.gamma.pdf(X, gamma_a**2)
    
    # center the maximum to prevent data shift in time
    # This is achieved by padding the left/right of the kernel with zeros.
    mx=np.argmax(kernel)
    if mx<len(kernel)/2:
        kernel=np.append(np.zeros(len(kernel)-2*mx), kernel)
    elif mx>len(kernel)/2:
        kernel=np.append(kernel, np.zeros(mx-(len(kernel)-mx)))
    assert len(kernel)<padarr.shape[axis],\
        'The kernel is longer than the array to convolved, you must decrease sd.'

    # normalize kernel to prevent vertical scaling
    kernel=kernel/sum(kernel)


    ## Convolve array with kernel
    sarr = np.apply_along_axis(lambda m:np.convolve(m, kernel, mode='same'), axis=axis, arr=padarr)


    ## Remove padding
    sarr = sarr[slice_along_axis(C,-C,axis=axis)]
    assert np.all(sarr.shape==arr.shape)


    return sarr


def slice_along_axis(a,b,s=1,axis=0):
    """
    Returns properly formatted slice to slice array/list along specified axis.
    - a: start
    - b: end
    - s: step
    """
    slc = slice(a,b,s)
    return (slice(None),) * axis + (slc,)


def xcorr_axis(x, y, axis=0):
    """
    Cross-correlation between two Nd arrays
    only along the specified axis
    (eg. for 2D array along axis 1, equivalent of applying np.correlate() to every pair of 1D array).

    Written because np.correlate does not handle axis argument, whereas fft does.

    Equivalent to (for axis 0, is x is 2D):
    c = np.zeros(x.shape)
    for i, (x_, y_) in enumerate(zip(normalize(x,0).T, normalize(y,0).T)):
        c[:,i] = np.correlate(x_, y_, mode='same')
    (actually this for loop is faster for small array because everything is done in C by numpy)
    """
    assert x.shape[axis]>3, f'array too short along axis {axis} to compute crosscorrelation.'
    y_rev = np.flip(y, axis=axis)
    c = irfft(rfft(x, axis=axis)*rfft(y_rev, axis=axis), axis=axis)

    # now need to simply reindex along the specified axis
    c_first = c.take(indices=range(0, c.shape[axis]//2-1), axis=axis)
    c_last = c.take(indices=range(c.shape[axis]//2-1, c.shape[axis]), axis=axis)

    return np.concatenate([c_last, c_first], axis=axis)

def xcorr_1d_fft(w1, w2, axis = 0):
    """
    Cross-correlation along specific axis between two Nd arrays.
    Typically for template matching when you know that you only need
    to 'slide' the template along a single axis (e.g. a column of 100x3 pixels in a 100x100 image,
    or a waveform across channels (10x82 samples) and a waveform template (10x82 as well)).

    w1, w2: Nd arrays
    axis: int, axis along which to perform crosscorrelation
    """
    c = xcorr_axis(normalize(w1,axis), normalize(w2,axis), axis=axis)
    # reindex along axis - fucked up at the moment
    return c

def xcorr_1d_loop(w1, w2):
    """
    Cross-correlation along axis 0 between two 2D arrays.
    Typically for template matching when you know that you only need
    to 'slide' the template along a single axis (e.g. a column of 100x3 pixels in a 100x100 image,
    or a waveform across channels (10x82 samples) and a waveform template (10x82 as well)).

    Faster than xcorr_1d_fft for small 2D arrays.

    w1, w2: 2d arrays
    """
    c = np.zeros(w1.shape)
    for i, (wave1, wave2) in enumerate(zip(normalize(w1,0).T, normalize(w2,0).T)):
        c[:,i] = np.correlate(wave1, wave2, mode='same')
    c[np.isnan(c)]=0 # replace nans with 0 correlation
    return c

def xcorr_2d(w1, w2):
    """
    Cross-correlation along ALL axis between two Nd arrays.
    Typically for template matching when you want to 'slide' the template along all axis
    (e.g. a 10x10 element pixels in a 100x100 image,
    or a 10x10x10 voxel in a 100x100x100 volume).
    """
    c = sgnl.correlate(normalize(w1, 0), normalize(w2, 0),  mode='same')
    return c

def normalize(x, axis=0):
    """
    Vanilla normalization (center, reduce) along specified axis
    """
    s = np.std(x, axis=axis)
    m = np.mean(x, axis=axis)
    return (x-m)/s

def get_bins(cwin, cbin):
    mod2=(cwin/cbin)%2
    if mod2==0: # even
        return np.arange(-cwin/2, cwin/2+cbin, cbin)
    else: # odd
        return np.arange(-cwin/2+cbin/2, cwin/2+cbin/2, cbin)

def find_nearest(array, value):
    array = np.asarray(array)
    idx = (np.abs(array - value)).argmin()
    return idx

def mask_2d(x, m):
    '''
    Mask a 2D array and preserve the
    dimension on the resulting array
    ----------
    x: np.array
       2D array on which to apply a mask
    m: np.array
        2D boolean mask
    Returns
    -------
    List of arrays. Each array contains the
    elements from the rows in x once masked.
    If no elements in a row are selected the
    corresponding array will be empty
    '''
    take = m.sum(axis=1)
    return np.split(x[m], np.cumsum(take)[:-1])

@njit(cache=True)
def make_2D_array(arr_lis, accept_heterogeneous=False):
    """Function to get 2D array from a list of lists
    """
    lis=[np.asarray(l) for l in arr_lis]
    n = len(lis)
    lengths = [len(l) for l in lis]
    if accept_heterogeneous:
        max_len = max(lengths)
        arr = np.zeros((n, max_len))*np.nan
        for i,l in enumerate(lis):
            arr[i, :lengths[i]] = l
    else:
        assert max(lengths)==min(lengths)
        arr = np.zeros((n, len(lis[0])))*np.nan
        for i,l in enumerate(lis):
            arr[i, :] = l
    return arr

@njit(cache=True)
def split(arr, sample_size=0, n_samples=0, overlap=0, return_last=True, verbose=True):
    '''
    Parameters:
        - arr: array to split into EITHER n_samples OR samples of size sample_size.
               samples_size has priority over n_samples.
        - s_samples: int, size of of samples to split the array into.
        - n_samples: int, number of samples to split the array into.
        - overlap: 0<=float<1, fraction of overlapping between consecutive samples | Default 0
        - return_last: bool, whether to return the last sample (whose size usually does not match the ones before).
    Returns:
        samples: array or list of samples
    '''

    arr=np.asarray(arr)
    assert n_samples!=0 or sample_size!=0, 'You need to specify either a sample number or size!'
    assert 0<=overlap<1, 'overlap needs to be between 0 and 1, 1 excluded (samples would all be the same)!'

    if n_samples!=0 and sample_size!=0:
        print('''WARNING you provided n_samples AND sample_size!
              By convention, sample_size has priority over n_samples. n_samples ignored.''')
        n_samples=0

    if n_samples==0: n_samples=len(arr)//sample_size+1
    if sample_size==0: sample_size=len(arr)//(n_samples)

    assert n_samples<=len(arr)

    # step=1 for maximum overlap (every sample would be the first sample for step = 0);
    # step=s for 0 overlap; overlap is
    s=sample_size
    if len(arr)<s: # no split necessary
        return make_2D_array([list(arr)])
    step = s-round(s*overlap)
    real_o=round((s-step)/s, 2)
    if overlap!=real_o and verbose: print('Real overlap: ', round(real_o, 2))
    samples = List([arr[i : i + s] for i in range(0, len(arr), step)])

    # always return last sample if len matches
    if len(samples[-1])==s:
        return make_2D_array(samples)

    if return_last:
        return make_2D_array(samples, accept_heterogeneous=True)

    lasti=-1
    sp=samples[lasti]
    while len(sp)!=s:
        lasti-=1
        sp=samples[lasti]
    return make_2D_array(samples[:lasti+1])

# def n_largest_samples(to_sort: np.array, largest_n: int) -> np.array:

#     """
#     Returns the n largest sorted samples from an array
#     """

#     sorted_n = np.argpartition(to_sort, -largest_n, axis=0 )[-largest_n:].flatten()
#     return sorted_n

def align_timeseries(timeseries, sync_signals, fs, offset_policy='original'):
    '''
    Usage 1: align >=2 time series in the same temporal reference frame with the same sampling frequency fs
        aligned_ts1, aligned_ts2, ... = align_timeseries([ts1,ts2,...], [sync1,sync2,...], fs)
    Usage 2:  align 1 time serie to another temporal reference frame
        aligned_ts = align_timeseries([ts], [sync_ts, sync_other], [fs_ts, fs_other])

    Re-aligns in time series based on provided sync signals.
    - timeseries: list of numpy arrays of time stamps (e.g. spikes), in SAMPLES
      If Usage 1: THEY MUST BE IN THE SAME TIME REFERENCE FRAME
    - sync_signals: list of numpy arrays of synchronization time stamps,
      ordered with respect to timeseries
      If Usage 1: THEY MUST ALSO BE IN THE SAME TIME REFERENCE FRAME
      - fs: int (usage 1) or list of 2 ints (usage 2), sampling frequencies of timeseries and respective sync_signals.
      - offset_policy: 'original' or 'zero', only for usage 1: whether to set timeseries[0] as 0 or as its original value after alignement.
      The FIRST sync_signal is used as a reference.

    Returns:
     timeseries, aligned accordingly to sync signals.

    '''
    assert type(timeseries) is type(sync_signals) is list, "You must provide timeseries and sync_signals as lists of arrays!"
    for tsi, ts in enumerate(timeseries):
        assert np.all(ts.astype(np.int64)==ts), 'Timeseries need to be integers, in samples acquired at fs sampling rate!'
        timeseries[tsi]=ts.astype(np.int64)
    for tsi, ts in enumerate(sync_signals):
        assert np.all(ts.astype(np.int64)==ts), 'Sync signals need to be integers, in samples acquired at fs sampling rate!'
        sync_signals[tsi]=ts.astype(np.int64)


    for ss in sync_signals:
        assert len(sync_signals[0])==len(ss), "WARNING all sync signals do not have the same size, the acquisition must have been faulty!"
    assert len(sync_signals[0])>=1, "Only one synchronization signal has been provided - this is dangerous practice as this does not account for cumulative time drift."
    if len(sync_signals[0])>50:
        print('More than 50 sync signals found - for performance reasons, sub-sampling to 50 homogenoeously spaced sync signals to align data.')
        subselect_ids=np.random.choice(np.arange(1,len(sync_signals[0])-1), 48, replace=False)
        subselect_ids=np.unique(np.append(subselect_ids,[0,len(sync_signals[0])-1])) # enforce first and last stim
        for synci,sync in enumerate(sync_signals):
            sync_signals[synci]=sync[subselect_ids]

    if len(timeseries)==1: # Usage 2
        usage=2
        offset=sync_signals[1][0]
        assert len(sync_signals)==2 and len(npa([fs]).flatten())==2, '''When providing a single time series ts, you need to provide 2 sync signals,
                                        [sync_ts, sync_other] to align ts to sync_other reference frame!'''
        fs_master=fs[1]
        fsconv=fs[1]/fs[0]
        timeseries=[sync_signals[1], timeseries[0]*fsconv]
        sync_signals=[sync_signals[1], sync_signals[0]*fsconv]
    elif len(timeseries)>=2: # Usage 1
        usage=1
        assert offset_policy in ['zero', 'original'], "offset_policy must be in ['zero', 'original']"
        offset = 0 if offset_policy=='original' else -timeseries[0][0]
        assert len(timeseries)==len(sync_signals)>=2, "There must be as many time series as sync signals, at least 2 of each!"
        assert len(npa([fs]).flatten())==1, 'You must provide a single sampling frequency fs when aligning >=2 time series (Usage 1)!'
        fs_master=npa([fs]).flatten()[0]


    Nevents, totDft, avDft, stdDft = len(sync_signals[0]), (sync_signals[1]-sync_signals[0])[-1], np.mean(np.diff(sync_signals[1]-sync_signals[0])), np.std(np.diff(sync_signals[1]-sync_signals[0]))
    totDft, avDft, stdDft = totDft*1000/fs_master, avDft*1000/fs_master, stdDft*1000/fs_master
    print("{} sync events used for alignement - start-end drift of {}ms".format(Nevents, round(totDft,3)))

    for dataset_i in range(len(timeseries)):
        if dataset_i==0: continue #first dataset is reference so left untouched
        array0, syncs = timeseries[dataset_i], sync_signals[dataset_i]
        array=array0-syncs[0] # initially center on first sync
        for i, sync in enumerate(syncs):
            if i>0: # first sync already subtracted
                array1=array0-sync # refer to own sync stamp, sync
                # re-time to where own sync stamp should be with respect to reference dataset (the 1st one)
                first_sync_ref0=sync_signals[0][i]-sync_signals[0][0]
                array=np.append(array[array1<0], array1[array1>=0]+first_sync_ref0)
        if usage==1: array+=syncs[0] # restore original offset
        timeseries[dataset_i]=array+offset

    return timeseries if usage==1 else timeseries[1]


def  align_timeseries_interpol(timeseries, sync_signals, fs=None):
    '''
    Align a list of N timeseries in the temporal reference frame of the first timeserie.

    Assumes that the drift is going to be linear, it is interpolated for times far from sync signals
    (sync_signal1 = a * sync_signal0 + b).

    Parameters:
    - timeseries: list[array[int]], list of np arrays (len N), timeseries to align. In SAMPLES to ensure accurate alignment
    - sync_signals: list[array[int]], list of np arrays (len N), sync signals respective to timeseries. In SAMPLES to ensure accurate alignment
    fs: float or list of floats (len N), sampling frequencies of time series.
        fs is optional (only to print drift and offset in seconds).

    returns:
        - timeseries: list of np arrays (len N, in samples), timeries aligned in temporal reference frame of timeseries[0]
    '''

    # Parameters formatting and checks
    assert len(timeseries)>=2
    assert len(sync_signals)==len(timeseries)
    if fs is not None:
        if assert_iterable(fs):assert len(timeseries)==len(fs)
        else:fs=[fs]*len(timeseries)
    for tsi, ts in enumerate(timeseries):
        assert np.all(ts.astype(np.int64)==ts), 'Timeseries need to be integers, in samples acquired at fs sampling rate!'
        timeseries[tsi]=ts.astype(np.int64)
    for tsi, ts in enumerate(sync_signals):
        assert np.all(ts.astype(np.int64)==ts), 'Sync signals need to be integers, in samples acquired at fs sampling rate!'
        sync_signals[tsi]=ts.astype(np.int64)

    # Align
    ref_sync=sync_signals[0]

    for i, (ts, ss) in enumerate(zip(timeseries[1:], sync_signals[1:])):
        (a, b) = np.polyfit(ss, ref_sync, 1)
        if fs is not None:
            drift=round(abs(a*fs[i]/fs[0]-1)*3600*1000,2)
            offset=round(b/fs[0],2)
            print(f'Drift (assumed linear) of {drift}ms/h, \noffset of {offset}s between time series 1 and {i+2}.\n')
        timeseries[i+1]=np.round(a*ts+b, 0).astype(np.int64)

    return timeseries

#%% Stolen from phy
def _as_array(arr, dtype=None):
    """Convert an object to a numerical NumPy array.
    Avoid a copy if possible.
    """
    if arr is None:
        return None
    if isinstance(arr, np.ndarray) and dtype is None:
        return arr
    if isinstance(arr, integer_types + (float,)):
        arr = [arr]
    out = np.asarray(arr)
    if dtype is not None:
        if out.dtype != dtype:
            out = out.astype(dtype)
    if out.dtype not in _ACCEPTED_ARRAY_DTYPES:
        raise ValueError("'arr' seems to have an invalid dtype: "
                         "{0:s}".format(str(out.dtype)))
    return out

def _unique(x):
    """Faster version of np.unique().
    This version is restricted to 1D arrays of non-negative integers.
    It is only faster if len(x) >> len(unique(x)).
    """
    if x is None or len(x) == 0:
        return np.array([], dtype=np.int64)
    # WARNING: only keep positive values.
    # cluster=-1 means "unclustered".
    x = _as_array(x)
    x = x[x >= 0]
    bc = np.bincount(x)
    return np.nonzero(bc)[0]

def _index_of(arr, lookup):
    """Replace scalars in an array by their indices in a lookup table.
    Implicitely assume that:
    * All elements of arr and lookup are non-negative integers.
    * All elements or arr belong to lookup.
    This is not checked for performance reasons.
    """
    # Equivalent of np.digitize(arr, lookup) - 1, but much faster.
    lookup = np.asarray(lookup, dtype=np.int32)
    m = (lookup.max() if len(lookup) else 0) + 1
    tmp = np.zeros(m + 1, dtype=np.int)
    # Ensure that -1 values are kept.
    tmp[-1] = -1
    if len(lookup):
        tmp[lookup] = np.arange(len(lookup))
    return tmp[arr]

#%% Peakdetect functions

__all__ = [
        "peakdetect",
        "peakdetect_fft",
        "peakdetect_parabola",
        "peakdetect_sine",
        "peakdetect_sine_locked",
        "peakdetect_spline",
        "peakdetect_zero_crossing",
        "zero_crossings",
        "zero_crossings_sine_fit"
        ]



def _datacheck_peakdetect(x_axis, y_axis):
    if x_axis is None:
        x_axis = range(len(y_axis))

    if len(y_axis) != len(x_axis):
        raise ValueError(
                "Input vectors y_axis and x_axis must have same length")

    #needs to be a numpy array
    y_axis = np.array(y_axis)
    x_axis = np.array(x_axis)
    return x_axis, y_axis


def _pad(fft_data, pad_len):
    """
    Pads fft data to interpolate in time domain

    keyword arguments:
    fft_data -- the fft
    pad_len --  By how many times the time resolution should be increased by

    return: padded list
    """
    l = len(fft_data)
    n = _n(l * pad_len)
    fft_data = list(fft_data)

    return fft_data[:l // 2] + [0] * (2**n-l) + fft_data[l // 2:]

def _n(x):
    """
    Find the smallest value for n, which fulfils 2**n >= x

    keyword arguments:
    x -- the value, which 2**n must surpass

    return: the integer n
    """
    return int(log(x)/log(2)) + 1


def _peakdetect_parabola_fitter(raw_peaks, x_axis, y_axis, points):
    """
    Performs the actual parabola fitting for the peakdetect_parabola function.

    keyword arguments:
    raw_peaks -- A list of either the maxima or the minima peaks, as given
        by the peakdetect functions, with index used as x-axis

    x_axis -- A numpy array of all the x values

    y_axis -- A numpy array of all the y values

    points -- How many points around the peak should be used during curve
        fitting, must be odd.


    return: A list giving all the peaks and the fitted waveform, format:
        [[x, y, [fitted_x, fitted_y]]]

    """
    func = lambda x, a, tau, c: a * ((x - tau) ** 2) + c
    fitted_peaks = []
    distance = abs(x_axis[raw_peaks[1][0]] - x_axis[raw_peaks[0][0]]) / 4
    for peak in raw_peaks:
        index = peak[0]
        x_data = x_axis[index - points // 2: index + points // 2 + 1]
        y_data = y_axis[index - points // 2: index + points // 2 + 1]
        # get a first approximation of tau (peak position in time)
        tau = x_axis[index]
        # get a first approximation of peak amplitude
        c = peak[1]
        a = np.sign(c) * (-1) * (np.sqrt(abs(c))/distance)**2
        """Derived from ABC formula to result in a solution where A=(rot(c)/t)**2"""

        # build list of approximations

        p0 = (a, tau, c)
        popt, pcov = curve_fit(func, x_data, y_data, p0)
        # retrieve tau and c i.e x and y value of peak
        x, y = popt[1:3]

        # create a high resolution data set for the fitted waveform
        x2 = np.linspace(x_data[0], x_data[-1], points * 10)
        y2 = func(x2, *popt)

        fitted_peaks.append([x, y, [x2, y2]])

    return fitted_peaks


def peakdetect_parabole(*args, **kwargs):
    """
    Misspelling of peakdetect_parabola
    function is deprecated please use peakdetect_parabola
    """
    logging.warn("peakdetect_parabole is deprecated due to misspelling use: peakdetect_parabola")

    return peakdetect_parabola(*args, **kwargs)


def peakdetect(y_axis, x_axis = None, lookahead = 200, delta=0):
    """
    Converted from/based on a MATLAB script at:
    http://billauer.co.il/peakdet.html

    function for detecting local maxima and minima in a signal.
    Discovers peaks by searching for values which are surrounded by lower
    or larger values for maxima and minima respectively

    keyword arguments:
    y_axis -- A list containing the signal over which to find peaks

    x_axis -- A x-axis whose values correspond to the y_axis list and is used
        in the return to specify the position of the peaks. If omitted an
        index of the y_axis is used.
        (default: None)

    lookahead -- distance to look ahead from a peak candidate to determine if
        it is the actual peak
        (default: 200)
        '(samples / period) / f' where '4 >= f >= 1.25' might be a good value

    delta -- this specifies a minimum difference between a peak and
        the following points, before a peak may be considered a peak. Useful
        to hinder the function from picking up false peaks towards to end of
        the signal. To work well delta should be set to delta >= RMSnoise * 5.
        (default: 0)
            When omitted delta function causes a 20% decrease in speed.
            When used Correctly it can double the speed of the function


    return: two lists [max_peaks, min_peaks] containing the positive and
        negative peaks respectively. Each cell of the lists contains a tuple
        of: (position, peak_value)
        to get the average peak value do: np.mean(max_peaks, 0)[1] on the
        results to unpack one of the lists into x, y coordinates do:
        x, y = zip(*max_peaks)
    """
    max_peaks = []
    min_peaks = []
    dump = []   #Used to pop the first hit which almost always is false

    # check input data
    x_axis, y_axis = _datacheck_peakdetect(x_axis, y_axis)
    # store data length for later use
    length = len(y_axis)


    #perform some checks
    if lookahead < 1:
        raise ValueError("Lookahead must be '1' or above in value")
    if not (np.isscalar(delta) and delta >= 0):
        raise ValueError("delta must be a positive number")

    #maxima and minima candidates are temporarily stored in
    #mx and mn respectively
    mn, mx = np.Inf, -np.Inf

    #Only detect peak if there is 'lookahead' amount of points after it
    for index, (x, y) in enumerate(zip(x_axis[:-lookahead],
                                        y_axis[:-lookahead])):
        if y > mx:
            mx = y
            mxpos = x
        if y < mn:
            mn = y
            mnpos = x

        ####look for max####
        if y < mx-delta and mx != np.Inf:
            #Maxima peak candidate found
            #look ahead in signal to ensure that this is a peak and not jitter
            if y_axis[index:index+lookahead].max() < mx:
                max_peaks.append([mxpos, mx])
                dump.append(True)
                #set algorithm to only find minima now
                mx = np.Inf
                mn = np.Inf
                if index+lookahead >= length:
                    #end is within lookahead no more peaks can be found
                    break
                continue
            #else:  #slows shit down this does
            #    mx = ahead
            #    mxpos = x_axis[np.where(y_axis[index:index+lookahead]==mx)]

        ####look for min####
        if y > mn+delta and mn != -np.Inf:
            #Minima peak candidate found
            #look ahead in signal to ensure that this is a peak and not jitter
            if y_axis[index:index+lookahead].min() > mn:
                min_peaks.append([mnpos, mn])
                dump.append(False)
                #set algorithm to only find maxima now
                mn = -np.Inf
                mx = -np.Inf
                if index+lookahead >= length:
                    #end is within lookahead no more peaks can be found
                    break
            #else:  #slows shit down this does
            #    mn = ahead
            #    mnpos = x_axis[np.where(y_axis[index:index+lookahead]==mn)]


    #Remove the false hit on the first value of the y_axis
    try:
        if dump[0]:
            max_peaks.pop(0)
        else:
            min_peaks.pop(0)
        del dump
    except IndexError:
        #no peaks were found, should the function return empty lists?
        pass

    return [max_peaks, min_peaks]


def peakdetect_fft(y_axis, x_axis, pad_len = 20):
    """
    Performs a FFT calculation on the data and zero-pads the results to
    increase the time domain resolution after performing the inverse fft and
    send the data to the 'peakdetect' function for peak
    detection.

    Omitting the x_axis is forbidden as it would make the resulting x_axis
    value silly if it was returned as the index 50.234 or similar.

    Will find at least 1 less peak then the 'peakdetect_zero_crossing'
    function, but should result in a more precise value of the peak as
    resolution has been increased. Some peaks are lost in an attempt to
    minimize spectral leakage by calculating the fft between two zero
    crossings for n amount of signal periods.

    The biggest time eater in this function is the ifft and thereafter it's
    the 'peakdetect' function which takes only half the time of the ifft.
    Speed improvements could include to check if 2**n points could be used for
    fft and ifft or change the 'peakdetect' to the 'peakdetect_zero_crossing',
    which is maybe 10 times faster than 'peakdetct'. The pro of 'peakdetect'
    is that it results in one less lost peak. It should also be noted that the
    time used by the ifft function can change greatly depending on the input.

    keyword arguments:
    y_axis -- A list containing the signal over which to find peaks

    x_axis -- A x-axis whose values correspond to the y_axis list and is used
        in the return to specify the position of the peaks.

    pad_len -- By how many times the time resolution should be
        increased by, e.g. 1 doubles the resolution. The amount is rounded up
        to the nearest 2**n amount
        (default: 20)


    return: two lists [max_peaks, min_peaks] containing the positive and
        negative peaks respectively. Each cell of the lists contains a tuple
        of: (position, peak_value)
        to get the average peak value do: np.mean(max_peaks, 0)[1] on the
        results to unpack one of the lists into x, y coordinates do:
        x, y = zip(*max_peaks)
    """
    # check input data
    x_axis, y_axis = _datacheck_peakdetect(x_axis, y_axis)
    zero_indices = zero_crossings(y_axis, window_len = 11)
    #select a n amount of periods
    last_indice = - 1 - (1 - len(zero_indices) & 1)
    ###
    # Calculate the fft between the first and last zero crossing
    # this method could be ignored if the beginning and the end of the signal
    # are unnecessary as any errors induced from not using whole periods
    # should mainly manifest in the beginning and the end of the signal, but
    # not in the rest of the signal
    # this is also unnecessary if the given data is an amount of whole periods
    ###
    fft_data = fft(y_axis[zero_indices[0]:zero_indices[last_indice]])
    padd = lambda x, c: x[:len(x) // 2] + [0] * c + x[len(x) // 2:]
    n = lambda x: int(log(x)/log(2)) + 1
    # pads to 2**n amount of samples
    fft_padded = padd(list(fft_data), 2 **
                n(len(fft_data) * pad_len) - len(fft_data))

    # There is amplitude decrease directly proportional to the sample increase
    sf = len(fft_padded) / float(len(fft_data))
    # There might be a leakage giving the result an imaginary component
    # Return only the real component
    y_axis_ifft = ifft(fft_padded).real * sf #(pad_len + 1)
    x_axis_ifft = np.linspace(
                x_axis[zero_indices[0]], x_axis[zero_indices[last_indice]],
                len(y_axis_ifft))
    # get the peaks to the interpolated waveform
    max_peaks, min_peaks = peakdetect(y_axis_ifft, x_axis_ifft, 500,
                                    delta = abs(np.diff(y_axis).max() * 2))
    #max_peaks, min_peaks = peakdetect_zero_crossing(y_axis_ifft, x_axis_ifft)

    # store one 20th of a period as waveform data
    data_len = int(np.diff(zero_indices).mean()) / 10
    data_len += 1 - data_len & 1


    return [max_peaks, min_peaks]


def peakdetect_parabola(y_axis, x_axis, points = 31):
    """
    Function for detecting local maxima and minima in a signal.
    Discovers peaks by fitting the model function: y = k (x - tau) ** 2 + m
    to the peaks. The amount of points used in the fitting is set by the
    points argument.

    Omitting the x_axis is forbidden as it would make the resulting x_axis
    value silly, if it was returned as index 50.234 or similar.

    will find the same amount of peaks as the 'peakdetect_zero_crossing'
    function, but might result in a more precise value of the peak.

    keyword arguments:
    y_axis -- A list containing the signal over which to find peaks

    x_axis -- A x-axis whose values correspond to the y_axis list and is used
        in the return to specify the position of the peaks.

    points -- How many points around the peak should be used during curve
        fitting (default: 31)


    return: two lists [max_peaks, min_peaks] containing the positive and
        negative peaks respectively. Each cell of the lists contains a tuple
        of: (position, peak_value)
        to get the average peak value do: np.mean(max_peaks, 0)[1] on the
        results to unpack one of the lists into x, y coordinates do:
        x, y = zip(*max_peaks)
    """
    # check input data
    x_axis, y_axis = _datacheck_peakdetect(x_axis, y_axis)
    # make the points argument odd
    points += 1 - points % 2
    #points += 1 - int(points) & 1 slower when int conversion needed

    # get raw peaks
    max_raw, min_raw = peakdetect_zero_crossing(y_axis)

    # define output variable
    max_peaks = []
    min_peaks = []

    max_ = _peakdetect_parabola_fitter(max_raw, x_axis, y_axis, points)
    min_ = _peakdetect_parabola_fitter(min_raw, x_axis, y_axis, points)

    max_peaks = map(lambda x: [x[0], x[1]], max_)
    max_fitted = map(lambda x: x[-1], max_)
    min_peaks = map(lambda x: [x[0], x[1]], min_)
    min_fitted = map(lambda x: x[-1], min_)

    return [max_peaks, min_peaks]


def peakdetect_sine(y_axis, x_axis, points = 31, lock_frequency = False):
    """
    Function for detecting local maxima and minima in a signal.
    Discovers peaks by fitting the model function:
    y = A * sin(2 * pi * f * (x - tau)) to the peaks. The amount of points used
    in the fitting is set by the points argument.

    Omitting the x_axis is forbidden as it would make the resulting x_axis
    value silly if it was returned as index 50.234 or similar.

    will find the same amount of peaks as the 'peakdetect_zero_crossing'
    function, but might result in a more precise value of the peak.

    The function might have some problems if the sine wave has a
    non-negligible total angle i.e. a k*x component, as this messes with the
    internal offset calculation of the peaks, might be fixed by fitting a
    y = k * x + m function to the peaks for offset calculation.

    keyword arguments:
    y_axis -- A list containing the signal over which to find peaks

    x_axis -- A x-axis whose values correspond to the y_axis list and is used
        in the return to specify the position of the peaks.

    points -- How many points around the peak should be used during curve
        fitting (default: 31)

    lock_frequency -- Specifies if the frequency argument of the model
        function should be locked to the value calculated from the raw peaks
        or if optimization process may tinker with it.
        (default: False)


    return: two lists [max_peaks, min_peaks] containing the positive and
        negative peaks respectively. Each cell of the lists contains a tuple
        of: (position, peak_value)
        to get the average peak value do: np.mean(max_peaks, 0)[1] on the
        results to unpack one of the lists into x, y coordinates do:
        x, y = zip(*max_peaks)
    """
    # check input data
    x_axis, y_axis = _datacheck_peakdetect(x_axis, y_axis)
    # make the points argument odd
    points += 1 - points % 2
    #points += 1 - int(points) & 1 slower when int conversion needed

    # get raw peaks
    max_raw, min_raw = peakdetect_zero_crossing(y_axis)

    # define output variable
    max_peaks = []
    min_peaks = []

    # get global offset
    offset = np.mean([np.mean(max_raw, 0)[1], np.mean(min_raw, 0)[1]])
    # fitting a k * x + m function to the peaks might be better
    #offset_func = lambda x, k, m: k * x + m

    # calculate an approximate frequency of the signal
    Hz_h_peak = np.diff(zip(*max_raw)[0]).mean()
    Hz_l_peak = np.diff(zip(*min_raw)[0]).mean()
    Hz = 1 / np.mean([Hz_h_peak, Hz_l_peak])



    # model function
    # if cosine is used then tau could equal the x position of the peak
    # if sine were to be used then tau would be the first zero crossing
    if lock_frequency:
        func = lambda x_ax, A, tau: A * np.sin(
            2 * pi * Hz * (x_ax - tau) + pi / 2)
    else:
        func = lambda x_ax, A, Hz, tau: A * np.sin(
            2 * pi * Hz * (x_ax - tau) + pi / 2)
    #func = lambda x_ax, A, Hz, tau: A * np.cos(2 * pi * Hz * (x_ax - tau))


    #get peaks
    fitted_peaks = []
    for raw_peaks in [max_raw, min_raw]:
        peak_data = []
        for peak in raw_peaks:
            index = peak[0]
            x_data = x_axis[index - points // 2: index + points // 2 + 1]
            y_data = y_axis[index - points // 2: index + points // 2 + 1]
            # get a first approximation of tau (peak position in time)
            tau = x_axis[index]
            # get a first approximation of peak amplitude
            A = peak[1]

            # build list of approximations
            if lock_frequency:
                p0 = (A, tau)
            else:
                p0 = (A, Hz, tau)

            # subtract offset from wave-shape
            y_data -= offset
            popt, pcov = curve_fit(func, x_data, y_data, p0)
            # retrieve tau and A i.e x and y value of peak
            x = popt[-1]
            y = popt[0]

            # create a high resolution data set for the fitted waveform
            x2 = np.linspace(x_data[0], x_data[-1], points * 10)
            y2 = func(x2, *popt)

            # add the offset to the results
            y += offset
            y2 += offset
            y_data += offset

            peak_data.append([x, y, [x2, y2]])

        fitted_peaks.append(peak_data)

    # structure date for output
    max_peaks = map(lambda x: [x[0], x[1]], fitted_peaks[0])
    max_fitted = map(lambda x: x[-1], fitted_peaks[0])
    min_peaks = map(lambda x: [x[0], x[1]], fitted_peaks[1])
    min_fitted = map(lambda x: x[-1], fitted_peaks[1])


    return [max_peaks, min_peaks]


def peakdetect_sine_locked(y_axis, x_axis, points = 31):
    """
    Convenience function for calling the 'peakdetect_sine' function with
    the lock_frequency argument as True.

    keyword arguments:
    y_axis -- A list containing the signal over which to find peaks
    x_axis -- A x-axis whose values correspond to the y_axis list and is used
        in the return to specify the position of the peaks.
    points -- How many points around the peak should be used during curve
        fitting (default: 31)

    return: see the function 'peakdetect_sine'
    """
    return peakdetect_sine(y_axis, x_axis, points, True)


def peakdetect_spline(y_axis, x_axis, pad_len=20):
    """
    Performs a b-spline interpolation on the data to increase resolution and
    send the data to the 'peakdetect_zero_crossing' function for peak
    detection.

    Omitting the x_axis is forbidden as it would make the resulting x_axis
    value silly if it was returned as the index 50.234 or similar.

    will find the same amount of peaks as the 'peakdetect_zero_crossing'
    function, but might result in a more precise value of the peak.

    keyword arguments:
    y_axis -- A list containing the signal over which to find peaks

    x_axis -- A x-axis whose values correspond to the y_axis list and is used
        in the return to specify the position of the peaks.
        x-axis must be equally spaced.

    pad_len -- By how many times the time resolution should be increased by,
        e.g. 1 doubles the resolution.
        (default: 20)


    return: two lists [max_peaks, min_peaks] containing the positive and
        negative peaks respectively. Each cell of the lists contains a tuple
        of: (position, peak_value)
        to get the average peak value do: np.mean(max_peaks, 0)[1] on the
        results to unpack one of the lists into x, y coordinates do:
        x, y = zip(*max_peaks)
    """
    # check input data
    x_axis, y_axis = _datacheck_peakdetect(x_axis, y_axis)
    # could perform a check if x_axis is equally spaced
    #if np.std(np.diff(x_axis)) > 1e-15: raise ValueError
    # perform spline interpolations
    dx = x_axis[1] - x_axis[0]
    x_interpolated = np.linspace(x_axis.min(), x_axis.max(), len(x_axis) * (pad_len + 1))
    cj = cspline1d(y_axis)
    y_interpolated = cspline1d_eval(cj, x_interpolated, dx=dx,x0=x_axis[0])
    # get peaks
    max_peaks, min_peaks = peakdetect_zero_crossing(y_interpolated, x_interpolated)

    return [max_peaks, min_peaks]

def peakdetect_zero_crossing(y_axis, x_axis = None, window = 11):
    """
    Function for detecting local maxima and minima in a signal.
    Discovers peaks by dividing the signal into bins and retrieving the
    maximum and minimum value of each the even and odd bins respectively.
    Division into bins is performed by smoothing the curve and finding the
    zero crossings.

    Suitable for repeatable signals, where some noise is tolerated. Executes
    faster than 'peakdetect', although this function will break if the offset
    of the signal is too large. It should also be noted that the first and
    last peak will probably not be found, as this function only can find peaks
    between the first and last zero crossing.

    keyword arguments:
    y_axis -- A list containing the signal over which to find peaks

    x_axis -- A x-axis whose values correspond to the y_axis list
        and is used in the return to specify the position of the peaks. If
        omitted an index of the y_axis is used.
        (default: None)

    window -- the dimension of the smoothing window; should be an odd integer
        (default: 11)


    return: two lists [max_peaks, min_peaks] containing the positive and
        negative peaks respectively. Each cell of the lists contains a tuple
        of: (position, peak_value)
        to get the average peak value do: np.mean(max_peaks, 0)[1] on the
        results to unpack one of the lists into x, y coordinates do:
        x, y = zip(*max_peaks)
    """
    # check input data
    x_axis, y_axis = _datacheck_peakdetect(x_axis, y_axis)

    zero_indices = zero_crossings(y_axis, window_len = window)
    period_lengths = np.diff(zero_indices)

    bins_y = [y_axis[index:index + diff] for index, diff in
        zip(zero_indices, period_lengths)]
    bins_x = [x_axis[index:index + diff] for index, diff in
        zip(zero_indices, period_lengths)]

    even_bins_y = bins_y[::2]
    odd_bins_y = bins_y[1::2]
    even_bins_x = bins_x[::2]
    odd_bins_x = bins_x[1::2]
    hi_peaks_x = []
    lo_peaks_x = []

    #check if even bin contains maxima
    if abs(even_bins_y[0].max()) > abs(even_bins_y[0].min()):
        hi_peaks = [bin.max() for bin in even_bins_y]
        lo_peaks = [bin.min() for bin in odd_bins_y]
        # get x values for peak
        for bin_x, bin_y, peak in zip(even_bins_x, even_bins_y, hi_peaks):
            hi_peaks_x.append(bin_x[np.where(bin_y==peak)[0][0]])
        for bin_x, bin_y, peak in zip(odd_bins_x, odd_bins_y, lo_peaks):
            lo_peaks_x.append(bin_x[np.where(bin_y==peak)[0][0]])
    else:
        hi_peaks = [bin.max() for bin in odd_bins_y]
        lo_peaks = [bin.min() for bin in even_bins_y]
        # get x values for peak
        for bin_x, bin_y, peak in zip(odd_bins_x, odd_bins_y, hi_peaks):
            hi_peaks_x.append(bin_x[np.where(bin_y==peak)[0][0]])
        for bin_x, bin_y, peak in zip(even_bins_x, even_bins_y, lo_peaks):
            lo_peaks_x.append(bin_x[np.where(bin_y==peak)[0][0]])

    max_peaks = [[x, y] for x,y in zip(hi_peaks_x, hi_peaks)]
    min_peaks = [[x, y] for x,y in zip(lo_peaks_x, lo_peaks)]

    return [max_peaks, min_peaks]


def _smooth(x, window_len=11, window="hanning"):
    """
    smooth the data using a window of the requested size.

    This method is based on the convolution of a scaled window on the signal.
    The signal is prepared by introducing reflected copies of the signal
    (with the window size) in both ends so that transient parts are minimized
    in the beginning and end part of the output signal.

    keyword arguments:
    x -- the input signal

    window_len -- the dimension of the smoothing window; should be an odd
        integer (default: 11)

    window -- the type of window from 'flat', 'hanning', 'hamming',
        'bartlett', 'blackman', where flat is a moving average
        (default: 'hanning')

    return: the smoothed signal

    example:
    t = linspace(-2,2,0.1)
    x = sin(t)+randn(len(t))*0.1
    y = _smooth(x)

    see also:

    numpy.hanning, numpy.hamming, numpy.bartlett, numpy.blackman,
    numpy.convolve, scipy.signal.lfilter
    """
    if x.ndim != 1:
        raise ValueError("smooth only accepts 1 dimension arrays.")

    if x.size < window_len:
        raise ValueError("Input vector needs to be bigger than window size.")

    if window_len<3:
        return x
    #declare valid windows in a dictionary
    window_funcs = {
        "flat": lambda _len: np.ones(_len, "d"),
        "hanning": np.hanning,
        "hamming": np.hamming,
        "bartlett": np.bartlett,
        "blackman": np.blackman
        }

    s = np.r_[x[window_len-1:0:-1], x, x[-1:-window_len:-1]]
    try:
        w = window_funcs[window](window_len)
    except KeyError:
        raise ValueError(
            "Window is not one of '{0}', '{1}', '{2}', '{3}', '{4}'".format(
            *window_funcs.keys()))

    y = np.convolve(w / w.sum(), s, mode = "valid")

    return y


def zero_crossings(y_axis, window_len = 11,
    window_f="hanning", offset_corrected=False):
    """
    Algorithm to find zero crossings. Smooths the curve and finds the
    zero-crossings by looking for a sign change.


    keyword arguments:
    y_axis -- A list containing the signal over which to find zero-crossings

    window_len -- the dimension of the smoothing window; should be an odd
        integer (default: 11)

    window_f -- the type of window from 'flat', 'hanning', 'hamming',
        'bartlett', 'blackman' (default: 'hanning')

    offset_corrected -- Used for recursive calling to remove offset when needed


    return: the index for each zero-crossing
    """
    # smooth the curve
    length = len(y_axis)

    # discard tail of smoothed signal
    y_axis = _smooth(y_axis, window_len, window_f)[:length]
    indices = np.where(np.diff(np.sign(y_axis)))[0]

    # check if zero-crossings are valid
    diff = np.diff(indices)
    if diff.std() / diff.mean() > 0.1:
        #Possibly bad zero crossing, see if it's offsets
        if ((diff[::2].std() / diff[::2].mean()) < 0.1 and
        (diff[1::2].std() / diff[1::2].mean()) < 0.1 and
        not offset_corrected):
            #offset present attempt to correct by subtracting the average
            offset = np.mean([y_axis.max(), y_axis.min()])
            return zero_crossings(y_axis-offset, window_len, window_f, True)
        #Invalid zero crossings and the offset has been removed
        print(diff.std() / diff.mean())
        print(np.diff(indices))
        raise ValueError(
            "False zero-crossings found, indicates problem {0!s} or {1!s}".format(
            "with smoothing window", "unhandled problem with offset"))
    # check if any zero crossings were found
    if len(indices) < 1:
        raise ValueError("No zero crossings found")
    #remove offset from indices due to filter function when returning
    return indices - (window_len // 2 - 1)
    # used this to test the fft function's sensitivity to spectral leakage
    #return indices + np.asarray(30 * np.random.randn(len(indices)), int)

############################Frequency calculation#############################
#    diff = np.diff(indices)
#    time_p_period = diff.mean()
#
#    if diff.std() / time_p_period > 0.1:
#        raise ValueError(
#            "smoothing window too small, false zero-crossing found")
#
#    #return frequency
#    return 1.0 / time_p_period
##############################################################################


def zero_crossings_sine_fit(y_axis, x_axis, fit_window = None, smooth_window = 11):
    """
    Detects the zero crossings of a signal by fitting a sine model function
    around the zero crossings:
    y = A * sin(2 * pi * Hz * (x - tau)) + k * x + m
    Only tau (the zero crossing) is varied during fitting.

    Offset and a linear drift of offset is accounted for by fitting a linear
    function the negative respective positive raw peaks of the wave-shape and
    the amplitude is calculated using data from the offset calculation i.e.
    the 'm' constant from the negative peaks is subtracted from the positive
    one to obtain amplitude.

    Frequency is calculated using the mean time between raw peaks.

    Algorithm seems to be sensitive to first guess e.g. a large smooth_window
    will give an error in the results.


    keyword arguments:
    y_axis -- A list containing the signal over which to find peaks

    x_axis -- A x-axis whose values correspond to the y_axis list
        and is used in the return to specify the position of the peaks. If
        omitted an index of the y_axis is used. (default: None)

    fit_window -- Number of points around the approximate zero crossing that
        should be used when fitting the sine wave. Must be small enough that
        no other zero crossing will be seen. If set to none then the mean
        distance between zero crossings will be used (default: None)

    smooth_window -- the dimension of the smoothing window; should be an odd
        integer (default: 11)


    return: A list containing the positions of all the zero crossings.
    """
    # check input data
    x_axis, y_axis = _datacheck_peakdetect(x_axis, y_axis)
    #get first guess
    zero_indices = zero_crossings(y_axis, window_len = smooth_window)
    #modify fit_window to show distance per direction
    if fit_window == None:
        fit_window = np.diff(zero_indices).mean() // 3
    else:
        fit_window = fit_window // 2

    #x_axis is a np array, use the indices to get a subset with zero crossings
    approx_crossings = x_axis[zero_indices]



    #get raw peaks for calculation of offsets and frequency
    raw_peaks = peakdetect_zero_crossing(y_axis, x_axis)
    #Use mean time between peaks for frequency
    ext = lambda x: list(zip(*x)[0])
    _diff = map(np.diff, map(ext, raw_peaks))


    Hz = 1 / np.mean(map(np.mean, _diff))
    #Hz = 1 / np.diff(approx_crossings).mean() #probably bad precision


    #offset model function
    offset_func = lambda x, k, m: k * x + m
    k = []
    m = []
    amplitude = []

    for peaks in raw_peaks:
        #get peak data as nparray
        x_data, y_data = map(np.asarray, zip(*peaks))
        #x_data = np.asarray(x_data)
        #y_data = np.asarray(y_data)
        #calc first guess
        A = np.mean(y_data)
        p0 = (0, A)
        popt, pcov = curve_fit(offset_func, x_data, y_data, p0)
        #append results
        k.append(popt[0])
        m.append(popt[1])
        amplitude.append(abs(A))

    #store offset constants
    p_offset = (np.mean(k), np.mean(m))
    A = m[0] - m[1]
    #define model function to fit to zero crossing
    #y = A * sin(2*pi * Hz * (x - tau)) + k * x + m
    func = lambda x, tau: A * np.sin(2 * pi * Hz * (x - tau)) + offset_func(x, *p_offset)


    #get true crossings
    true_crossings = []
    for indice, crossing in zip(zero_indices, approx_crossings):
        p0 = (crossing, )
        subset_start = max(indice - fit_window, 0.0)
        subset_end = min(indice + fit_window + 1, len(x_axis) - 1.0)
        x_subset = np.asarray(x_axis[subset_start:subset_end])
        y_subset = np.asarray(y_axis[subset_start:subset_end])
        #fit
        popt, pcov = curve_fit(func, x_subset, y_subset, p0)

        true_crossings.append(popt[0])


    return true_crossings



# -*- coding: utf-8 -*-
"""
Created on Fri May  3 16:30:50 2019

@author: Maxime Beau, Hausser lab, University College London

Input/output utilitaries to deal with Neuropixels files.
"""

import psutil
import shutil
import os
from ast import literal_eval as ale
from pathlib import Path
from tqdm.auto import tqdm

from math import ceil
import numpy as np
try:
    import cupy as cp
except ImportError:
    print(("cupy could not be imported - "
    "some functions dealing with the binary file (filtering, whitening...) will not work."))

from npyx.utils import npa, read_pyfile, list_files
from npyx.preprocess import apply_filter, bandpass_filter, whitening, approximated_whitening_matrix, med_substract,\
                            gpufilter, cu_median, adc_realign, kfilt

import json

#%% Load metadata and channel map

def read_metadata(dp):
    f'''
    {metadata.__doc__}

    If ran on a merged dataset, an additional layer of keys is added: the probes used as keys of
    dp_dict when npyx.merger.merge_datasets was ran (typically 'probe1' and 'probe2'):
    the structure of meta is then 'probe1':meta_data_dataset1,
                                  'probe2':meta_data_dataset_2, ...

    '''

    if assert_multi(dp):
        meta = {}
        for dpx, probe in get_ds_table(dp).loc[:,'dp':'probe'].values:
            meta[probe] = metadata(dpx)
    else:
        meta = metadata(dp)

    return meta


def metadata(dp):
    '''
    Read spikeGLX (.ap/lf.meta) or openEphys (.oebin) metadata files
    and returns their contents as dictionnaries.

    The 'highpass' or 'lowpass' nested dicts correspond to Neuropixels 1.0 high or low pass filtered metadata.
    2.0 recordings only have a 'highpass' key, as they are acquired as a single file matched with a .ap.meta file.
        for spikeGLX, corresponds to metadata of .ap.meta and .lf.meta files.
        for OpenEphys, .oebin metadata relating to the first and second dictionnaries in 'continuous' of the .oebin file
                       which match the /continuous/Neuropix-PXI-100.0 or .1 folders respectively.

    Parameters:
        - dp: str, datapath to spike sorted dataset

    Returns:
        - meta: dictionnary containing contents of meta file.
        the structure of meta is as follow:
        {
        'probe_version':'3A', '1.0_staggered', '2.0_1shank', '2.0_4shanks',
        'highpass':
            {
            'binary_relative_path':relative path to binary file from dp,
            'sampling_rate':int, # sampling rate
            'n_channels_binaryfile':int, # n channels saved on file, typically 385 for .bin and 384 for .dat
            'n_channels_analysed':int, # n channels used for spikesorting. Will set the shape of temp_wh.daat for kilosort.
            'datatype':str, # datatype of binary encoding, typically int16
            'binary_relative_path':relative path to binary file from dp,
            'key1...': all other keys present in meta file, that you must be familiar with!
                       e.g. 'fileSizeBytes' for spikeGLX or 'channels' for OpenEphys...
            },
        'lowpass': {...}, # same as high for low pass filtered data (not existing in 2.0 recordings)
        'events': {...}, # only for openephys recordings, contents of oebin file
        'spikes': {...} # only for openephys recordings, contents of oebin file
        }
    '''
    dp = Path(dp)
    assert dp.exists(), "Provided path does not exist!"
    assert dp.is_dir(), f"Provided path {dp} is a filename!"

    probe_versions = {
        'glx':{3.0:'3A', # option 3
               0.0:'1.0',
               21:'2.0_singleshank',
               24:'2.0_fourshanks'},
        'oe':{"Neuropix-3a":'3A', # source_processor_name keys
                "Neuropix-PXI":'1.0',
                '?1':'2.0_singleshank', # do not know yet
                '?2':'2.0_fourshanks'}, # do not know yet
        'int':{'3A':1, '1.0':1,
               '2.0_singleshank':2, '2.0_fourshanks':2}
        }

    # import params.py data
    params_f = dp/'params.py'
    if params_f.exists():
        params=read_pyfile(dp/'params.py')

    # find meta file
    glx_ap_files = list_files(dp, "ap.meta", True)
    glx_lf_files = list_files(dp, "lf.meta", True)
    oe_files = list_files(dp, "oebin", True)
    glx_found = np.any(glx_ap_files) or np.any(glx_lf_files)
    oe_found = np.any(oe_files)
    assert glx_found or oe_found, \
        f'WARNING no .ap/lf.meta (spikeGLX) or .oebin (OpenEphys) file found at {dp}.'
    assert not (glx_found and oe_found),\
        'WARNING dataset seems to contain both an open ephys and spikeGLX metafile - fix this!'
    assert len(glx_ap_files)==1 or len(glx_lf_files)==1 or len(oe_files)==1,\
        'WARNING more than 1 .ap.meta or 1 .oebin files found!'

    # Formatting of openephys meta file
    meta = {}
    meta['path'] = os.path.realpath(dp)
    if oe_found:
        meta['acquisition_software']='OpenEphys'
        # Load OpenEphys metadata
        metafile=Path(oe_files[0])
        with open(metafile) as f:
            meta_oe = json.load(f)

        # find probe version
        oe_probe_version = meta_oe["continuous"][0]["source_processor_name"]
        assert oe_probe_version in probe_versions['oe'].keys(),\
            f'WARNING only probe version {oe_probe_version} not handled with openEphys - post an issue at www.github.com/m-beau/NeuroPyxels'
        meta['probe_version']=probe_versions['oe'][oe_probe_version]
        meta['probe_version_int'] = probe_versions['int'][meta['probe_version']]

        # Find conversion factor
        # should be 0.19499999284744262695
        meta['bit_uV_conv_factor']=meta_oe["continuous"][0]["channels"][0]["bit_volts"]

        # find everything else
        for filt_key in ['highpass','lowpass']:
            meta[filt_key]={}
            filt_key_i={'highpass':0, 'lowpass':1}[filt_key]
            meta[filt_key]['sampling_rate']=int(meta_oe["continuous"][filt_key_i]['sample_rate'])
            meta[filt_key]['n_channels_binaryfile']=int(meta_oe["continuous"][filt_key_i]['num_channels'])
            if params_f.exists():
                meta[filt_key]['n_channels_analysed']=params['n_channels_dat']
                meta[filt_key]['datatype']=params['dtype']
            else:
                meta[filt_key]['n_channels_analysed']=meta[filt_key]['n_channels_binaryfile']
                meta[filt_key]['datatype']='int16'
            binary_folder = './continuous/'+meta_oe["continuous"][filt_key_i]['folder_name']
            binary_file = list_files(dp/binary_folder, "dat", False)
            if any(binary_file):
                binary_rel_path = binary_folder+binary_file[0]
                meta[filt_key]['binary_relative_path']=binary_rel_path
                meta[filt_key]['binary_byte_size']=os.path.getsize(dp/binary_rel_path)
                if filt_key=='highpass' and params_f.exists():
                    if params['dat_path']!=binary_rel_path:
                        print((f'\033[34;1mWARNING edit dat_path in params.py '
                        f'so that it matches relative location of high pass filtered binary file: {binary_rel_path}'))
            else:
                meta[filt_key]['binary_relative_path']='not_found'
                meta[filt_key]['binary_byte_size']='unknown'
                print(f"\033[91;1mWARNING {filt_key} binary file not found at {dp}\033[0m")
            meta[filt_key]={**meta[filt_key], **meta_oe["continuous"][filt_key_i]}
        meta["events"]=meta_oe["events"]
        meta["spikes"]=meta_oe["spikes"]


    # Formatting of SpikeGLX meta file
    elif glx_found:
        meta['acquisition_software']='SpikeGLX'
        # Load SpikeGLX metadata
        meta_glx = {}
        for metafile in glx_ap_files+glx_lf_files:
            if metafile in glx_ap_files: filtkey='highpass'
            elif metafile in glx_lf_files: filtkey='lowpass'
            metafile=Path(metafile)
            meta_glx[filtkey]={}
            with open(metafile, 'r') as f:
                for ln in f.readlines():
                    tmp = ln.split('=')
                    k, val = tmp[0], ''.join(tmp[1:])
                    k = k.strip()
                    val = val.strip('\r\n')
                    if '~' in k:
                        meta_glx[filtkey][k] = val.strip('(').strip(')').split(')(')
                    else:
                        try:  # is it numeric?
                            meta_glx[filtkey][k] = float(val)
                        except:
                            meta_glx[filtkey][k] = val

        # find probe version
        if 'imProbeOpt' in meta_glx["highpass"].keys(): # 3A
            glx_probe_version = meta_glx["highpass"]["imProbeOpt"]
        elif 'imDatPrb_type' in meta_glx["highpass"].keys(): # 1.0 and beyond
            glx_probe_version = meta_glx["highpass"]["imDatPrb_type"]
        else:
             glx_probe_version = 'N/A'

        assert glx_probe_version in probe_versions['glx'].keys(),\
            f'WARNING probe version {glx_probe_version} not handled - post an issue at www.github.com/m-beau/NeuroPyxels'
        meta['probe_version']=probe_versions['glx'][glx_probe_version]
        meta['probe_version_int'] = probe_versions['int'][meta['probe_version']]

        # Based on probe version,
        # Find the voltage range, gain, encoding
        # and deduce the conversion from units/bit to uV
        Vrange=(meta_glx["highpass"]['imAiRangeMax']-meta_glx["highpass"]['imAiRangeMin'])*1e6
        if meta['probe_version'] in ['3A', '1.0']:
            if Vrange!=1.2e6: print(f'\u001b[31mHeads-up, the voltage range seems to be {Vrange}, which is not the default (1.2*10^6). Might be normal!')
            bits_encoding=10
            ampFactor=ale(meta_glx["highpass"]['~imroTbl'][1].split(' ')[3]) # typically 500
            #if ampFactor!=500: print(f'\u001b[31mHeads-up, the voltage amplification factor seems to be {ampFactor}, which is not the default (500). Might be normal!')
        elif meta['probe_version'] in ['2.0_singleshank', '2.0_fourshanks']:
            if Vrange!=1e6: print(f'\u001b[31mHeads-up, the voltage range seems to be {Vrange}, which is not the default (10^6). Might be normal!')
            bits_encoding=14
            ampFactor=80 # hardcoded
        meta['bit_uV_conv_factor']=(Vrange/2**bits_encoding/ampFactor)


        # find everything else
        for filt_key in ['highpass','lowpass']:
            if filt_key not in meta_glx.keys(): continue
            meta[filt_key]={}

            # binary file
            filt_suffix={'highpass':'ap','lowpass':'lf'}[filt_key]
            binary_rel_path = get_binary_file_path(dp, filt_suffix, False)
            if binary_rel_path!='not_found':
                meta[filt_key]['binary_byte_size']=os.path.getsize(dp/binary_rel_path)
                meta[filt_key]['binary_relative_path']='./'+binary_rel_path
            else:
                meta[filt_key]['binary_byte_size']='unknown'
                meta[filt_key]['binary_relative_path']=binary_rel_path
                #print(f"\033[91;1mWARNING binary file .{filt_suffix}.bin not found at {dp}\033[0m")

            # sampling rate
            if meta_glx[filt_key]['typeThis'] == 'imec':
                meta[filt_key]['sampling_rate']=int(meta_glx[filt_key]['imSampRate'])
            else:
                meta[filt_key]['sampling_rate']=int(meta_glx[meta_glx['typeThis'][:2]+'SampRate'])

            meta[filt_key]['n_channels_binaryfile']=int(meta_glx[filt_key]['nSavedChans'])
            if params_f.exists():
                meta[filt_key]['n_channels_analysed']=params['n_channels_dat']
                meta[filt_key]['datatype']=params['dtype']
            else:
                meta[filt_key]['n_channels_analysed']=meta[filt_key]['n_channels_binaryfile']
                meta[filt_key]['datatype']='int16'
            meta[filt_key]={**meta[filt_key], **meta_glx[filt_key]}

    # Calculate length of recording
    high_fs = meta['highpass']['sampling_rate']
    if meta['highpass']['binary_byte_size']=='unknown':
        t_end=np.load(dp/'spike_times.npy').ravel()[-1]
        meta['recording_length_seconds']=t_end/high_fs
    else:
        file_size = meta['highpass']['binary_byte_size']
        item_size = np.dtype(meta['highpass']['datatype']).itemsize
        nChans = meta['highpass']['n_channels_binaryfile']
        meta['recording_length_seconds'] = file_size/item_size/nChans/high_fs


    return meta


def chan_map(dp=None, y_orig='surface', probe_version=None):
    '''
    Returns probe channel map.
    Parameters:
        - dp: str, datapath
        - y_orig: 'surface' or 'tip', where to position channel 0.
                If surface (default), channel map is flipped vertically (0 is now at the surface).
        - probe_version: None, 'local', '3A', '1.0' or '2.0_singleshank' (other types not handled yet, reach out to give your own!).
                        If 'local', will load channelmap from dp (only contains analyzed channels, not all channels)
                        If explicitely given, will return complete channelmap of electrode.
                        If None, will guess probe version from metadata and return complete channelmap.
    Returns:
        - chan_map: array of shape (N_electrodes, 3).
                    1st column is channel indices, 2nd x position, 3rd y position
    '''

    dp = Path(dp)
    assert y_orig in ['surface', 'tip']
    if probe_version is None: probe_version=read_metadata(dp)['probe_version']

    if probe_version in ['3A', '1.0']:
        Nchan=384
        cm_el = npa([[  43,   0],
                           [  11,   0],
                           [  59,   20],
                           [  27,   20]])
        vert=npa([[  0,   40],
                  [  0,   40],
                  [  0,   40],
                  [  0,   40]])

        cm=cm_el.copy()
        for i in range(int(Nchan/cm_el.shape[0])-1):
            cm = np.vstack((cm, cm_el+vert*(i+1)))
        cm=np.hstack([np.arange(Nchan).reshape(Nchan,1), cm])

    elif probe_version=='1.0':
        Nchan=384
        cm_el = npa([[  43,   0],
                     [  11,   0]])
        vert=npa([[  0,   20],
                  [  0,   20]])

        cm=cm_el.copy()
        for i in range(int(Nchan/cm_el.shape[0])-1):
            cm = np.vstack((cm, cm_el+vert*(i+1)))
        cm=np.hstack([np.arange(Nchan).reshape(Nchan,1), cm])

    elif probe_version=='2.0_singleshank':
        Nchan=384
        cm_el = npa([[  32,   0],
                     [  0,   0]])
        vert=npa([[  0,   15],
                  [  0,   15]])

        cm=cm_el.copy()
        for i in range(int(Nchan/cm_el.shape[0])-1):
            cm = np.vstack((cm, cm_el+vert*(i+1)))
        cm=np.hstack([np.arange(Nchan).reshape(Nchan,1), cm])

    else:
        probe_version='local'

    if probe_version=='local':
        if dp is None:
            raise ValueError("dp argument is not provided - when channel map is \
                             atypical and probe_version is hence called 'local', \
                             the datapath needs to be provided to load the channel map.")
        c_ind=np.load(dp/'channel_map.npy');cp=np.load(dp/'channel_positions.npy')
        cm=npa(np.hstack([c_ind.reshape(max(c_ind.shape),1), cp]), dtype=np.int32)

    if y_orig=='surface':
        cm[:,1:]=cm[:,1:][::-1]

    return cm

#%% Binary file I/O, including sync channel

def get_binary_file_path(dp, filt_suffix='ap', absolute_path=True):
    f'''Return the path of the binary file (.bin) from a directory.

    Wrapper of get_glx_file_path:
    {get_glx_file_path.__doc__}
    '''
    
    return get_glx_file_path(dp, 'bin', filt_suffix, absolute_path)

def get_meta_file_path(dp, filt_suffix='ap', absolute_path=True):
    f'''Return the path of the meta file (.meta) from a directory.

    Wrapper of get_glx_file_path:
    {get_glx_file_path.__doc__}
    '''
    
    return get_glx_file_path(dp, 'meta', filt_suffix, absolute_path)

def get_glx_file_path(dp, suffix, filt_suffix='ap', absolute_path=True):
    '''Return the path of a spikeGLX file (.bin or .meta) from a directory.

    Parameters:
    - dp: str, directory
    - filt_suffix: 'ap' or 'lf', seek ap (highpass) or lfp (lowpass) binary file,
                   for 1.0 recordings. Always 'ap' for 2.0.
    - absolute_path: bool, whether to return path from root
                     (if False, returns relative path from dp)

    Returns:
        The absolute path to the spikeGLX file with the associated filt key, 'ap' or 'lf'.
    '''

    dp = Path(dp)
    assert suffix in ['bin', 'meta']
    assert filt_suffix in ['ap','lf']
    glx_files = list_files(dp, f"{filt_suffix}.{suffix}", absolute_path)
    assert len(glx_files) <= 1, (f"More than one {filt_suffix}.{suffix} files found at {dp}! ",
        "If you keep several versions, store other files in a subdirectory (e.g. original_data).")

    if len(glx_files)==0:
        glx_files = ['not_found']
    
    return glx_files[0]

def unpackbits(x,num_bits = 16):
    '''
    unpacks numbers in bits.
    '''
    xshape = list(x.shape)
    x = x.reshape([-1,1])
    to_and = 2**np.arange(num_bits).reshape([1,num_bits])
    return (x & to_and).astype(bool).astype(np.int64).reshape(xshape + [num_bits])

def get_npix_sync(dp, output_binary = False, filt_key='highpass', unit='seconds',
                  verbose=False, again=False):
    '''Unpacks neuropixels external input data, to align spikes to events.
    Parameters:
        - dp: str, datapath
        - output_binary: bool, whether to output binary sync channel as 0/1s
        - filt_key: str, 'highpass' or 'lowpass' (SpikeGLX: ap/lf, OIpenEphys: Neuropix-PXI-100.0/.1)
        - unit: str, 'seconds' or 'samples', units of returned onsets/offset times
        - verbose: bool, whether to print rich information
        - again: bool, whether to reload sync channel from binary file.

    Returns:
        Dictionnaries of length n_channels = number of channels where threshold crossings were found, [0-16]
        - onsets: dict, {channel_i:np.array(onset1, onset2, ...), ...} in 'unit'
        - offsets: dict, {channel_i:np.array(offset1, offset2, ...), ...} in 'unit'

    '''
    dp = Path(dp)
    if assert_multi(dp):
        dp=Path(get_ds_table(dp)['dp'][0])
        if verbose: print(f'Loading npix sync channel from a merged dataset - assuming temporal reference frame of dataset 0:\n{dp}')

    assert filt_key in ['highpass', 'lowpass']
    filt_suffix = {'highpass':'ap', 'lowpass':'lf'}[filt_key]
    assert unit in ['seconds', 'samples']
    fname=''
    onsets={}
    offsets={}
    sync_dp=dp/'sync_chan'
    meta = read_metadata(dp)
    srate = meta[filt_key]['sampling_rate'] if unit=='seconds' else 1

    if meta['acquisition_software']=='OpenEphys':
        raise('OpenEphys sync channel loading not implemented yet - manually load ./Neuropix-PXI-100.0/TTL_1/timestamps.npy.')
        filt_id = 0 if filt_key=='highpass' else 1
        timestamps = np.load(dp/f'events/Neuropix-PXI-100.{filt_id}/TTL_1/timestamps.npy')

        onsets={ok:ov/srate for ok, ov in onsets.items()}
        offsets={ok:ov/srate for ok, ov in offsets.items()}

        return onsets,offsets

    elif meta['acquisition_software']=='SpikeGLX':

        # Tries to load pre-saved onsets and offsets
        if sync_dp.exists() and not output_binary:
            if verbose: print(f'Sync channel extraction directory found: {sync_dp}\n')
            for file in os.listdir(sync_dp):
                if file.endswith("on_samples.npy"):
                    filt_suffix_loaded=file.split('.')[-2][:2]
                    if filt_suffix_loaded==filt_suffix: # if samples are at the instructed sampling rate i.e. lf (2500) or ap (30000)!
                        if verbose: print(f'Sync channel onsets ({file}) file found and loaded.')
                        file_i = ale(file[-15])
                        onsets[file_i]=np.load(sync_dp/file)/srate
                elif file.endswith("of_samples.npy"):
                    filt_suffix_loaded=file.split('.')[-2][:2]
                    if filt_suffix_loaded==filt_suffix: # if samples are at the instructed sampling rate i.e. lf (2500) or ap (30000)!
                        if verbose: print(f'Sync channel offsets ({file}) file found and loaded.')
                        file_i = ale(file[-15])
                        offsets[file_i]=np.load(sync_dp/file)/srate
            if any(onsets):
                # else, might be that sync_dp is empty
                return onsets, offsets

        # Tries to load pre-saved compressed binary
        if sync_dp.exists() and not again:
            if verbose: print(f"No file ending in 'on_samples.npy' with the right sampling rate ({filt_suffix}) found in sync_chan directory: extracting sync channel from binary.\n")
            npz_files = list_files(sync_dp, 'npz')
            if any(npz_files):
                if verbose: print('Compressed binary found - extracting from there...')
                fname=npz_files[0][:-9]
                sync_fname = npz_files[0][:-4]
                binary=np.load(sync_dp/(sync_fname+'.npz'))
                binary=binary[dir(binary.f)[0]].astype(np.int8)

        else: os.mkdir(sync_dp)

        # If still no file name, memorymaps binary directly
        if fname=='':
            ap_files = list_files(dp, 'ap.bin')
            lf_files = list_files(dp, 'lf.bin')

            if filt_suffix=='ap':
                assert any(ap_files), f'No .ap.bin file found at {dp}!! Aborting.'
                fname=ap_files[0]
            elif filt_suffix=='lf':
                assert any(lf_files), f'No .lf.bin file found at {dp}!! Aborting.'
                fname=lf_files[0]

            nchan=meta[filt_key]['n_channels_binaryfile']
            dt=np.dtype(meta[filt_key]['datatype'])
            nsamples = os.path.getsize(dp/fname) / (nchan * dt.itemsize)
            syncdat=np.memmap(dp/fname,
                            mode='r',
                            dtype=dt,
                            shape=(int(nsamples), int(nchan)))[:,-1]


            print('Unpacking {}...'.format(fname))
            binary = unpackbits(syncdat.flatten(),16).astype(np.int8)
            sync_fname = fname[:-4]+'_sync'
            np.savez_compressed(sync_dp/(sync_fname+'.npz'), binary)

        if output_binary:
            return binary

        # Generates onsets and offsets from binary
        mult = 1
        sync_idx_onset = np.where(mult*np.diff(binary, axis = 0)>0)
        sync_idx_offset = np.where(mult*np.diff(binary, axis = 0)<0)
        for ichan in np.unique(sync_idx_onset[1]):
            ons = sync_idx_onset[0][
                sync_idx_onset[1] == ichan]
            onsets[ichan] = ons
            np.save(Path(sync_dp, sync_fname+'{}on_samples.npy'.format(ichan)), ons)
        for ichan in np.unique(sync_idx_offset[1]):
            ofs = sync_idx_offset[0][
                sync_idx_offset[1] == ichan]
            offsets[ichan] = ofs
            np.save(Path(sync_dp, sync_fname+'{}of_samples.npy'.format(ichan)), ofs)

        onsets={ok:ov/srate for ok, ov in onsets.items()}
        offsets={ok:ov/srate for ok, ov in offsets.items()}

        assert any(onsets), ("WARNING no sync channel found in dataset - "
            "make sure you are running this function on a dataset with a synchronization TTL!")

        return onsets,offsets

def extract_rawChunk(dp, times, channels=np.arange(384), filt_key='highpass', save=0,
                     whiten=0, med_sub=0, hpfilt=0, hpfiltf=300, nRangeWhiten=None, nRangeMedSub=None,
                     ignore_ks_chanfilt=0, center_chans_on_0=False, verbose=False, scale=True, again=False):
    '''Function to extract a chunk of raw data on a given range of channels on a given time window.
    ## PARAMETERS
    - dp: datapath to folder with binary path (files must ends in .bin, typically ap.bin)
    - times: list of boundaries of the time window, in seconds [t1, t2].
    - channels (default: np.arange(384)): list of channels of interest, in 0 indexed integers [c1, c2, c3...]
    - filt_key: 'ap' or 'lf', whether to exxtract from the high-pass or low-pass filtered binary file
    - save (default 0): save the raw chunk in the bdp directory as '{bdp}_t1-t2_c1-c2.npy'
    - whiten: whether to whiten the data across channels. If nRangeWhiten is not None, whitening matrix is computed with the nRangeWhiten closest channels.
    - med_sub: whether to median-subtract the data across channels. If nRangeMedSub is not none, median of each channel is computed using the nRangeMedSub closest channels.
    - hpfilt: whether to high-pass filter the data, using a 3 nodes butterworth filter of cutoff frequency hpfiltf.
    - hpfiltf: see hpfilt
    - nRangeWhiten: int, see whiten.
    - nRangeMedSub: int, see med_sub.
    - ignore_ks_chanfilt: whether to ignore the filtering made by kilosort, which only uses channels with average events rate > ops.minfr to spike sort. | Default False
    - scale: A boolean variable specifying whether we should convert the resulting raw
             A2D samples to uV. Defaults to True
    ## RETURNS
    rawChunk: numpy array of shape ((c2-c1), (t2-t1)*fs).
    rawChunk[0,:] is channel 0; rawChunk[1,:] is channel 1, etc.
    '''
    # Find binary file
    dp = Path(dp)
    meta = read_metadata(dp)
    fname = get_binary_file_path(dp, filt_suffix='ap', absolute_path=True)

    fs = meta[filt_key]['sampling_rate']
    Nchans=meta[filt_key]['n_channels_binaryfile']
    bytes_per_sample=2
    whitenpad=200

    assert len(times)==2
    assert times[0]>=0
    assert times[1]<meta['recording_length_seconds']

    # Format inputs
    cm=chan_map(dp, probe_version='local'); assert cm.shape[0]<=Nchans-1
    if not ignore_ks_chanfilt: channels=assert_chan_in_dataset(dp, channels) # index out of 384, should remain the same because rc initial shape is 384!
    t1, t2 = int(np.round(times[0]*fs)), int(np.round(times[1]*fs))
    if whiten:
        if t1<whitenpad:
            print(f"times[0] set to {round(whitenpad/30000, 5)}s because whitening requires a pad.")
            t1 = whitenpad
            times[0] = t1/30000
        t1, t2 = t1-whitenpad, t2+whitenpad
    bn = os.path.basename(fname) # binary name
    rcn = f'{bn}_t{times[0]}-{times[1]}_ch{channels[0]}-{channels[-1]}_{whiten}_{med_sub}_{scale}.npy' # raw chunk name
    rcp = get_npyx_memory(dp) / rcn

    if os.path.isfile(rcp) and not again:
        return np.load(rcp)

    # Check that available memory is high enough to load the raw chunk
    vmem=dict(psutil.virtual_memory()._asdict())
    chunkSize = int(fs*Nchans*bytes_per_sample*(times[1]-times[0]))
    if verbose:
        print('Used RAM: {0:.1f}% ({1:.2f}GB total).'.format(vmem['used']*100/vmem['total'], vmem['total']/1024/1024/1024))
        print('Chunk size:{0:.3f}MB. Available RAM: {1:.3f}MB.'.format(chunkSize/1024/1024, vmem['available']/1024/1024))
    if chunkSize>0.9*vmem['available']:
        print('WARNING you are trying to load {0:.3f}MB into RAM but have only {1:.3f}MB available.\
              Pick less channels or a smaller time chunk.'.format(chunkSize/1024/1024, vmem['available']/1024/1024))
        return

    # Get chunk from binary file
    with open(fname, 'rb') as f_src:
        # each sample for each channel is encoded on 16 bits = 2 bytes: samples*Nchannels*2.
        byte1 = int(t1*Nchans*bytes_per_sample)
        byte2 = int(t2*Nchans*bytes_per_sample)
        bytesRange = byte2-byte1

        f_src.seek(byte1)

        bData = f_src.read(bytesRange)

    # Decode binary data
    # channels on axis 0, time on axis 1
    assert len(bData)%2==0
    rc = np.frombuffer(bData, dtype=np.int16) # 16bits decoding
    rc = rc.reshape((int(t2-t1), Nchans)).T

    # Align signal on each channel
    if center_chans_on_0:
        rc=rc-np.median(rc[:,:10],axis=1)[:,np.newaxis]

    # Median subtraction = CAR
    if med_sub:
        rc=med_substract(rc, 0, nRange=nRangeMedSub)

    # Highpass filter with a 3rd order butterworth filter, like in kilosort2
    if hpfilt:
        rc=apply_filter(rc, bandpass_filter(rate=fs, low=None, high=hpfiltf, order=3), axis=1)

    # Whiten data
    if whiten:
        rc=whitening(rc, nRange=nRangeWhiten)
        rc=rc[:, whitenpad:-whitenpad]

    # get the right channels range, AFTER WHITENING
    rc = rc[channels, :]
    # Scale data
    if scale:
        rc = rc*meta['bit_uV_conv_factor'] # convert into uV

    # convert from cupy to numpy array
    if 'cp' in globals():
        rc = cp.asnumpy(rc)

    if save: # sync chan saved in extract_syncChan
        np.save(rcp, rc)

    return rc

def assert_chan_in_dataset(dp, channels):
    channels = np.array(channels)
    cm=chan_map(dp, probe_version='local')
    if not np.all(np.isin(channels, cm[:,0])):
        print(("WARNING Kilosort excluded some channels that you provided for analysis "
               "because they did not display enough threshold crossings! Jumping channels:"
               f"{channels[~np.isin(channels, cm[:,0])]}"))
    channels=channels[np.isin(channels, cm[:,0])]
    return channels

#%% Binary file filtering wrappers

def preprocess_binary_file(dp=None, filt_key='ap', fname=None, target_dp=None, move_orig_data=True,
                       ADC_realign = False, median_subtract=False, f_low=None, f_high=300, order=3,
                       spatial_filt=False, whiten = False, whiten_range=32,
                       again_Wrot=False, verbose=False):
    """Creates a preprocessed copy of binary file at dp/fname_filtered.bin,
    and moves the original binary file to dp/original_data.fname.bin.

    One must precise either dp (path to directory or ) or fname (absolute path to binary file).

    Preprocessing steps:
    - optional - realigning data according to ADCs, like global demux from CatGT does
    - optional - CAR, common average referencing (must be BEFORE filtering as filtering re-offsets each channel)
    - necessary - high pass filtering the data, to remove DC offsets
    - removing correlated noise across channels, using either/all of these 2 options:
        - spatial filtering with butterworth filter
        - whitening
        using both is not recommended, has not been well tested.

    By default, the data is ADC realigned -> CAR -> high pass filtered at 300Hz 
    with a 3nodes butterworth filter (bidirectional to prevent phase shifting).


    Parameters:
    - dp: optional str, path to binary file directory. dp/*.bin file will be found and used for filtering.
    - filt_key: str, 'ap' or 'lf' (if filtering ap.bin or lf.bin file)
    - fname: optional str, absolute path of binary file to filter (if provided, *.bin will not be guessed)
    - target_dp: str or Path, directory to save preprocessed binary file (by default, dp)
    - move_orig_data: bool, if true a directory is created at dp/original_data, and the original binary file is moved there.
    - ADC_realign: bool, whether to realign data based on Neuropixels ADC shifts (slow because requires FFT)
    - CAR: bool, whether to perform common average subtraction
    - f_low: optional float, lowpass filter frequency
    - f_high: float, highpass filter frequency (necessary)
    - order: int, butterworth filter order (default 3)
    - spatial_filt: bool, whether to high pass filter across channels at 0.1 Hz
    - whiten: bool, whether to whiten across channels
    - verbose: bool, whether to print extra information
    """

    # Parameters check
    assert dp is not None or fname is not None,\
        "You must either provide a path to the binary file directory (dp)\
            or the absolute path to the binary file (fname)."
    assert f_low is not None or f_high is not None,\
        "You must either provide a lowpass (low) or highpass (high) pass filter frequency."
    assert filt_key in ['ap', 'lf']

    # samples of symmetrical buffer for whitening and spike detection
    # Must be multiple of 32 + ntbuff. This is the batch size (try decreasing if out of memory).
    ntb = 64 # in kilosort: called ntbuff
    nSkipCov = 25

    # Fetch binary file name, define target file name
    if fname is None:
        fname = get_binary_file_path(dp, filt_key, True)
    fname=Path(fname)
    dp = fname.parent
    if target_dp is None:
        target_dp = dp
    else:
        target_dp = Path(target_dp)
    print(f"Preprocessing {fname}...")

    filter_suffix = ""
    message = ""
    if ADC_realign:
        filter_suffix+=f"_adcshift{ADC_realign}"
        message+="    - shifting ADCs,\n"
    if median_subtract:
        filter_suffix+="_medsub"
        message+="    - median subtraction (aka common average referencing CAR),\n"
    filter_suffix+=f"_tempfilt{f_low}{f_high}"
    low_s = 0 if f_low is None else f_low
    message+=f"    - filtering in time (between {low_s} and {f_high} Hz),\n"
    if whiten:
        filter_suffix+=f"_whit{whiten}{whiten_range}"
        message+=f"    - whitening (using {whiten_range} closest channels),\n"
    if spatial_filt:
        filter_suffix+=f"_spatfilt{spatial_filt}"
        message+=f"    - filtering in space ({spatial_filt} 'Hz'),\n"
    filtered_fname = str(fname.name)[:-7]+filter_suffix+".ap.bin"
    message = message[:-2]+"."
    print(message)

    # fetch metadata
    fk = {'ap':'highpass', 'lf':'lowpass'}[filt_key]
    meta = read_metadata(dp)
    fs = meta[fk]['sampling_rate']
    binary_byte_size = meta[fk]['binary_byte_size']

    # Check that there is enough free memory on disk
    free_memory = shutil.disk_usage(target_dp)[2]
    assert free_memory > binary_byte_size + 2**10,\
        f"Not enough free space on disk at {target_dp} (need {(binary_byte_size+2**10)//2**20} MB)"

    # memory map binary file
    n_channels = meta[fk]['n_channels_binaryfile']
    channels_to_process = np.arange(n_channels-1) # would allow in the future to process specific channels
    chans_mask = np.isin(np.arange(n_channels), channels_to_process)
    dtype = meta[fk]['datatype']
    offset = 0
    item_size = np.dtype(dtype).itemsize
    n_samples = (binary_byte_size - offset) // (item_size * n_channels)
    memmap_f = np.memmap(fname, dtype=dtype, offset=offset, shape=(n_samples, n_channels), mode='r+')


    # fetch whitening matrix (estimate covariance over a few batches)
    if whiten:
        Wrot_path = dp / 'whitening_matrix.npy'
        Wrot = approximated_whitening_matrix(memmap_f, Wrot_path, whiten_range,
            NT, Nbatch, NTbuff, ntb, nSkipCov, n_channels, channels_to_process,
            f_high, fs, again=again_Wrot, verbose=verbose)

    # Preprocess iteratively, batch by batch
    NT = 64 * 1024 + ntb
    Nbatch = ceil(n_samples / NT)
    NTbuff = NT + 3 * ntb
    w_edge = cp.linspace(0,1,ntb).reshape(-1, 1) # weights to combine data batches at the edge
    buff_prev = cp.zeros((ntb, n_channels-1), dtype=np.int32)
    last_batch=False
    with open(target_dp / filtered_fname, 'wb') as fw:  # open for writing processed data
        for ibatch in tqdm(range(Nbatch), desc="Preprocessing"):
            # we'll create a binary file of batches of NT samples, which overlap consecutively
            # on params.ntbuff samples
            # in addition to that, we'll read another params.ntbuff samples from before and after,
            # to have as buffers for filtering

            # Collect data batch
            i = max(0, NT * ibatch - ntb)
            rawData = memmap_f[i:i + NTbuff]
            if rawData.size == 0:
                print("Loaded buffer has an empty size!")
                break  # this shouldn't really happen, unless we counted data batches wrong
            nsampcurr = rawData.shape[0]  # how many time samples the current batch has
            if nsampcurr < NTbuff:
                # when reaching end of file, mirror end by adding missing samples to fit in GPU
                last_batch = True
                n_extra_samples = NTbuff - nsampcurr
                rawData = np.concatenate(
                    (rawData, np.tile(rawData[nsampcurr - 1], (n_extra_samples, 1))), axis=0)
            if i == 0:
                bpad = np.tile(rawData[0], (ntb, 1))
                rawData = np.concatenate((bpad, rawData[:NTbuff - ntb]), axis=0)
            rawData = cp.asarray(rawData, dtype=np.float32)

            # Extract channels to use for processing
            # at minima, removes sync channel
            batch = rawData[:,chans_mask]

            # Re-alignment based on ADCs shifts (like CatGT)
            # should be the first preprocessing step,
            # it simply consists in properly realigning the data!
            if ADC_realign:
                batch = cp.asnumpy(batch)
                batch = adc_realign(batch, version=meta['probe_version_int'])
                batch = cp.asarray(batch, dtype=np.float32)

            # CAR (optional) -> temporal filtering -> weight to combine edges -> unpadding
            batch = gpufilter(batch, fs=fs, fshigh=f_high, fslow=f_low, order=order,
                              car=median_subtract, bidirectional=False)
            assert batch.flags.c_contiguous # check that ordering is still C, not F
            batch[ntb:2*ntb] = w_edge * batch[ntb:2*ntb] + (1 - w_edge) * buff_prev
            buff_prev = batch[NT + ntb: NT + 2*ntb]
            batch = batch[ntb:ntb + NT, :]  # remove timepoints used as buffers

            # Spatial filtering (replaces whitening)
            if spatial_filt:
                batch = kfilt(batch.T, butter_kwargs = {'N': 3, 'Wn': 0.1, 'btype': 'highpass'}).T

            # whiten the data and scale by 200 for int16 range
            if whiten:
                print("Whitening not implemented yet.")
                batch = cp.dot(batch, Wrot)
            
            assert batch.flags.c_contiguous  # check that ordering is still C, not F
            if batch.shape[0] != NT:
                raise ValueError(f'Batch {ibatch} processed incorrectly')

            # add unprocessed channels back to batch
            # (at minima including last 16 bits for sync signal)
            rebuilt_batch = rawData[ntb:ntb + NT, :] # remove timepoints used as buffers; includes unprocessed channels
            rebuilt_batch[:,chans_mask] = batch
            if last_batch:
                # remove mirrored data at the end
                n_extra_samples = n_extra_samples - 2*ntb # account for buffers
                rebuilt_batch = rebuilt_batch[:-n_extra_samples]

            # convert to int16, and gather on the CPU side
            datcpu = cp.asnumpy(rebuilt_batch.astype(np.dtype(dtype)))

            # write this batch to binary file
            if verbose and (ibatch%(Nbatch//50)==0 or last_batch):
                print(f"{(target_dp/filtered_fname).stat().st_size} total, {rebuilt_batch.size * 2} bytes written to file {datcpu.shape} array size")
            datcpu.tofile(fw)
        if verbose: print(f"{(target_dp/filtered_fname).stat().st_size} total")

    # Finally, if everything ran smoothly,
    # move original binary to new directory
    if move_orig_data:
        orig_dp = dp/'original_data'
        orig_dp.mkdir(exist_ok=True)
        if not (orig_dp/fname.name).exists(): fname.replace(orig_dp/fname.name)
        meta_f = get_meta_file_path(dp, filt_key, False)
        if not (orig_dp/meta_f).exists():
            if (dp/meta_f).exists():
                shutil.copy(dp/meta_f, orig_dp/meta_f)
            if (dp/'channel_map.npy').exists():
                shutil.copy(dp/'channel_map.npy', orig_dp/'channel_map.npy')
            if (dp/'channel_positions.npy').exists():
                shutil.copy(dp/'channel_positions.npy', orig_dp/'channel_positions.npy')

    return target_dp/filtered_fname

#%% paqIO file loading utilities

def paq_read(file_path):
    """
    Read PAQ file (from PackIO) into python
    Lloyd Russell 2015
    See https://github.com/llerussell/paq2py

    Parameters
    ==========
    file_path : str, optional
        full path to file to read in.

    Returns
    =======
    data : ndarray
        the data as a m-by-n array where m is the number of channels and n is
        the number of datapoints
    chan_names : list of str
        the names of the channels provided in PackIO
    hw_chans : list of str
        the hardware lines corresponding to each channel
    units : list of str
        the units of measurement for each channel
    rate : int
        the acquisition sample rate, in Hz
    """

    # open file
    fid = open(file_path, 'rb')

    # get sample rate
    rate = int(np.fromfile(fid, dtype='>f', count=1))
    assert rate!=0, 'WARNING something went wrong with the paq file, redownload it.'

    # get number of channels
    num_chans = int(np.fromfile(fid, dtype='>f', count=1))

    # get channel names
    chan_names = []
    for i in range(num_chans):
        num_chars = int(np.fromfile(fid, dtype='>f', count=1))
        chan_name = ''
        for j in range(num_chars):
            chan_name = chan_name + chr(int(np.fromfile(fid, dtype='>f', count=1)))
        chan_names.append(chan_name)

    # get channel hardware lines
    hw_chans = []
    for i in range(num_chans):
        num_chars = int(np.fromfile(fid, dtype='>f', count=1))
        hw_chan = ''
        for j in range(num_chars):
            hw_chan = hw_chan + chr(int(np.fromfile(fid, dtype='>f', count=1)))
        hw_chans.append(hw_chan)

    # get acquisition units
    units = []
    for i in range(num_chans):
        num_chars = int(np.fromfile(fid, dtype='>f', count=1))
        unit = ''
        for j in range(num_chars):
            unit = unit + chr(int(np.fromfile(fid, dtype='>f', count=1)))
        units.append(unit)

    # get data
    temp_data = np.fromfile(fid, dtype='>f', count=-1)
    num_datapoints = int(len(temp_data)/num_chans)
    data = np.reshape(temp_data, [num_datapoints, num_chans]).transpose()

    # close file
    fid.close()

    return {"data": data,
            "chan_names": chan_names,
            "hw_chans": hw_chans,
            "units": units,
            "rate": rate}

#%% I/O array functions from phy

# def _start_stop(item):
#     """Find the start and stop indices of a __getitem__ item.

#     This is used only by ConcatenatedArrays.

#     Only two cases are supported currently:

#     * Single integer.
#     * Contiguous slice in the first dimension only.

#     """
#     if isinstance(item, tuple):
#         item = item[0]
#     if isinstance(item, slice):
#         # Slice.
#         if item.step not in (None, 1):
#             raise NotImplementedError()
#         return item.start, item.stop
#     elif isinstance(item, (list, np.ndarray)):
#         # List or array of indices.
#         return np.min(item), np.max(item)
#     else:
#         # Integer.
#         return item, item + 1

# def _fill_index(arr, item):
#     if isinstance(item, tuple):
#         item = (slice(None, None, None),) + item[1:]
#         return arr[item]
#     else:
#         return arr

# class ConcatenatedArrays(object):
#     """This object represents a concatenation of several memory-mapped
#     arrays. Coming from phy.io.array.py"""
#     def __init__(self, arrs, cols=None, scaling=None):
#         assert isinstance(arrs, list)
#         self.arrs = arrs
#         # Reordering of the columns.
#         self.cols = cols
#         self.offsets = np.concatenate([[0], np.cumsum([arr.shape[0]
#                                                        for arr in arrs])],
#                                       axis=0)
#         self.dtype = arrs[0].dtype if arrs else None
#         self.scaling = scaling

#     @property
#     def shape(self):
#         if self.arrs[0].ndim == 1:
#             return (self.offsets[-1],)
#         ncols = (len(self.cols) if self.cols is not None
#                  else self.arrs[0].shape[1])
#         return (self.offsets[-1], ncols)

#     def _get_recording(self, index):
#         """Return the recording that contains a given index."""
#         assert index >= 0
#         recs = np.nonzero((index - self.offsets[:-1]) >= 0)[0]
#         if len(recs) == 0:  # pragma: no cover
#             # If the index is greater than the total size,
#             # return the last recording.
#             return len(self.arrs) - 1
#         # Return the last recording such that the index is greater than
#         # its offset.
#         return recs[-1]

#     def _get(self, item):
#         cols = self.cols if self.cols is not None else slice(None, None, None)
#         # Get the start and stop indices of the requested item.
#         start, stop = _start_stop(item)
#         # Return the concatenation of all arrays.
#         if start is None and stop is None:
#             return np.concatenate(self.arrs, axis=0)[..., cols]
#         if start is None:
#             start = 0
#         if stop is None:
#             stop = self.offsets[-1]
#         if stop < 0:
#             stop = self.offsets[-1] + stop
#         # Get the recording indices of the first and last item.
#         rec_start = self._get_recording(start)
#         rec_stop = self._get_recording(stop)
#         assert 0 <= rec_start <= rec_stop < len(self.arrs)
#         # Find the start and stop relative to the arrays.
#         start_rel = start - self.offsets[rec_start]
#         stop_rel = stop - self.offsets[rec_stop]
#         # Single array case.
#         if rec_start == rec_stop:
#             # Apply the rest of the index.
#             out = _fill_index(self.arrs[rec_start][start_rel:stop_rel], item)
#             out = out[..., cols]
#             return out
#         chunk_start = self.arrs[rec_start][start_rel:]
#         chunk_stop = self.arrs[rec_stop][:stop_rel]
#         # Concatenate all chunks.
#         l = [chunk_start]
#         if rec_stop - rec_start >= 2:
#             print("Loading a full virtual array: this might be slow "
#                         "and something might be wrong.")
#             l += [self.arrs[r][...] for r in range(rec_start + 1,
#                                                    rec_stop)]
#         l += [chunk_stop]
#         # Apply the rest of the index.
#         return _fill_index(np.concatenate(l, axis=0), item)[..., cols]

#     def __getitem__(self, item):
#         out = self._get(item)
#         assert out is not None
#         if self.scaling is not None and self.scaling != 1:
#             out = out * self.scaling
#         return out

#     def __len__(self):
#         return self.shape[0]

# def _pad(arr, n, dir='right'):
#     """Pad an array with zeros along the first axis.

#     Parameters
#     ----------

#     n : int
#         Size of the returned array in the first axis.
#     dir : str
#         Direction of the padding. Must be one 'left' or 'right'.

#     """
#     assert dir in ('left', 'right')
#     if n < 0:
#         raise ValueError("'n' must be positive: {0}.".format(n))
#     elif n == 0:
#         return np.zeros((0,) + arr.shape[1:], dtype=arr.dtype)
#     n_arr = arr.shape[0]
#     shape = (n,) + arr.shape[1:]
#     if n_arr == n:
#         assert arr.shape == shape
#         return arr
#     elif n_arr < n:
#         out = np.zeros(shape, dtype=arr.dtype)
#         if dir == 'left':
#             out[-n_arr:, ...] = arr
#         elif dir == 'right':
#             out[:n_arr, ...] = arr
#         assert out.shape == shape
#         return out
#     else:
#         if dir == 'left':
#             out = arr[-n:, ...]
#         elif dir == 'right':
#             out = arr[:n, ...]
#         assert out.shape == shape
#         return out

# def _range_from_slice(myslice, start=None, stop=None, step=None, length=None):
#     """Convert a slice to an array of integers."""
#     assert isinstance(myslice, slice)
#     # Find 'step'.
#     step = myslice.step if myslice.step is not None else step
#     if step is None:
#         step = 1
#     # Find 'start'.
#     start = myslice.start if myslice.start is not None else start
#     if start is None:
#         start = 0
#     # Find 'stop' as a function of length if 'stop' is unspecified.
#     stop = myslice.stop if myslice.stop is not None else stop
#     if length is not None:
#         stop_inferred = floor(start + step * length)
#         if stop is not None and stop < stop_inferred:
#             raise ValueError("'stop' ({stop}) and ".format(stop=stop) +
#                              "'length' ({length}) ".format(length=length) +
#                              "are not compatible.")
#         stop = stop_inferred
#     if stop is None and length is None:
#         raise ValueError("'stop' and 'length' cannot be both unspecified.")
#     myrange = np.arange(start, stop, step)
#     # Check the length if it was specified.
#     if length is not None:
#         assert len(myrange) == length
#     return myrange



from npyx.gl import assert_multi, get_ds_table, get_npyx_memory
#



# -*- coding: utf-8 -*-
"""
2019-06-13
@author: Maxime Beau, Neural Computations Lab, University College London

Dataset merger class: represents a collection of Neuropixels Datasets (no theoretical upper limit on number of probes)
as a networkx network
with nodes being characterized as cell types
and edges as putative connections.

The way to use it is as follow:

# initiation
dp_dic = {'probe1': 'path/to/dataset1', 'probe2':'path/to/dataset2', ...}
merged = Merger(dp_dic)

# Connect the graph with putative monosynaptic connections
merged.connect_graph()

# Plot the graph
merged.plot_graph()

# Get N of putative connections of a given node spotted on the graph
merged.get_node_edges(node)

# Only keep a set of relevant nodes or edges and plot it again
merged.keep_edges(edges_list)
merged.keep_nodes_list(nodes_list)
merged.plot_graph()

# every graph operation of dataset merger can be performed on external networkx graphs
# provided with the argument 'src_graph'
g=merged.get_graph_copy(mergerGraph='undigraph')
merged.keep_nodes_list(nodes_list, src_graph=g) # g itself will be modified, not need to do g=...
merged.plot_graph(graph_src=g)


"""


import warnings

warnings.simplefilter(action="ignore", category=FutureWarning)

import os
import os.path as op

opj = op.join
from pathlib import Path

import numpy as np
import pandas as pd

from npyx.utils import npa, align_timeseries_interpol, assert_float

from npyx.inout import get_npix_sync
from npyx.gl import (
    get_units,
    load_merged_units_qualities,
    detect_new_spikesorting,
    save_qualities,
)


def merge_datasets(datapaths, again=False):
    """
    Merges datasets together and aligns data accross probes, modelling drift as a affine function.

    Any npyx function will run on a merged dataset just as on a regular dataset directory.

    If datasets have already been merged, will simply return the right datapath and dataset table.

    Parameters:
    - datapaths: dict of structure
        {'name_probe_1':'path/to/kilosort/output1',
         'name_probe_2':'path/to/kilosort/output2', ...}
    - again: bool, whether to force re-merge the datasets (even if a previous merge is found).

    Returns:
        - dp_merged: str, path to merged dataset
        - ds_table: pandas dataframe, containing datasets names, paths, probe names
    """

    # Handle datapaths format
    val_e = ValueError(
        """
    Datapath should be a dict of kilosort paths which exist:
    {'name_probe_1':'path/to/kilosort/output1', ...}"""
    )
    if type(datapaths) is dict:
        for i, v in datapaths.items():
            datapaths[i] = Path(v)
            if not datapaths[i].exists():
                raise val_e
    else:
        raise val_e

    # Make an alphabetically indexed table of all fed datasets (if several)
    ds_table = pd.DataFrame(columns=["dataset_name", "dp", "probe"])
    for i, (prb, dp) in enumerate(datapaths.items()):
        ds_table.loc[i, "dataset_name"] = op.basename(dp)
        ds_table.loc[i, "dp"] = str(dp)
        ds_table.loc[i, "probe"] = prb
    ds_table.insert(
        0, "dataset_i", ds_table["dataset_name"].argsort()
    )  # sort table by dataset_name alphabetically
    ds_table.sort_values("dataset_i", axis=0, inplace=True)
    ds_table.set_index("dataset_i", inplace=True)

    n_datasets = len(ds_table.index)
    mess_prefix = "\033[34;1m--- "
    mess_suffix = "\033[0m"
    if n_datasets == 1:
        print(
            f"{mess_prefix}Only one dataset provided - simply returning its path.{mess_suffix}."
        )
        return ds_table.loc[0, "dp"], ds_table

    # Instanciate datasets and define the merger path, where the 'merged' data will be saved,
    # together with the dataset table, holding information about the datasets (indices, names, source paths etc)
    ds_names = ""
    predirname = op.dirname(ds_table["dp"][0])
    for ds_i in ds_table.index:
        dp, prb = ds_table.loc[ds_i, "dp"], ds_table.loc[ds_i, "probe"]
        ds_names += "_" + ds_table.loc[ds_i, "dataset_name"]
        if predirname != op.dirname(dp):
            print(
                "WARNING: all your datasets are not stored in the same pre-directory - {} is picked anyway.".format(
                    predirname
                )
            )
    dp_merged = Path(predirname) / (
        "merged" + ds_names
    )  # Path(predirname, 'merger'+ds_names)
    old_prophyler_path = Path(predirname) / ("prophyler" + ds_names)
    if old_prophyler_path.exists():
        print(
            "Backward compatibility - prophyler renamed according to merger convention."
        )
        os.rename(str(old_prophyler_path), str(dp_merged))
    if not op.isdir(dp_merged):
        print(
            f"\n{mess_prefix}Merged data (from {n_datasets} datasets) will be saved here: {dp_merged}.{mess_suffix}"
        )
        os.mkdir(dp_merged)
    else:
        print(f"\n{mess_prefix}Merged dataset found at {dp_merged}.{mess_suffix}")
        ds_table_old = pd.read_csv(
            Path(dp_merged, "datasets_table.csv"), index_col="dataset_i"
        )
        old_dataset_names = list(ds_table_old.loc[:, "dataset_name"].values)
        for ds_i in ds_table.index:
            assert (
                ds_table.loc[ds_i, "dataset_name"] in old_dataset_names
            ), f"""WARNING you ran dataset merger on these {n_datasets} datasets in the past
                but used the dataset names {old_dataset_names}!!"""

    ds_table.to_csv(Path(dp_merged, "datasets_table.csv"))

    # Load and save units qualities
    # + check whether datasets have been re-spike sorted if not first instanciation
    qualities = load_merged_units_qualities(dp_merged, ds_table)
    re_spikesorted = detect_new_spikesorting(dp_merged, True, qualities)
    save_qualities(dp_merged, qualities)

    # Merge spike times (or copy them if only 1 dataset)
    # Only if merged_clusters_times does not exist already or does but re-spikesorting has been detected
    merge_fname_times = "spike_times"
    merge_fname_clusters = "spike_clusters"
    if (
        (not op.exists(Path(dp_merged, merge_fname_times + ".npy")))
        or re_spikesorted
        or again
    ):
        print(
            f"\n{mess_prefix}Loading spike trains of {n_datasets} datasets...{mess_suffix}"
        )
        # precompute all sync channels without prompting the user
        onsets = [
            get_npix_sync(dp, output_binary=False, filt_key="highpass", unit="samples")[
                0
            ]
            for dp in ds_table["dp"]
        ]
        spike_times, spike_clusters, sync_signals = [], [], []
        for ds_i, dp in enumerate(ds_table["dp"]):
            spike_times.append(np.load(Path(dp, "spike_times.npy")).flatten())
            spike_clusters.append(np.load(Path(dp, "spike_clusters.npy")).flatten())
            ons = onsets[ds_i]
            syncchan = ask_syncchan(ons)
            if syncchan == "q":
                print("Aborted. Returning Nones.")
                return None, None
            sync_signals.append(ons[syncchan])
        assert all(
            len(x) == len(sync_signals[0]) for x in sync_signals
        ), "WARNING different number of events on sync channels of both probes! Try again."
        NspikesTotal = 0
        for i in range(len(spike_times)):
            NspikesTotal += len(spike_times[i])

        # If several datasets are fed to the merger, align their spike times.
        # Merged dataset is saved as two arrays:
        # spike_times and spike_clusters, as in a regular dataset,
        # but where spike_clusters is a larger int (1 is unit 1 from dataset 0, 1,000,000,001 is unit 1 from dataset 1)
        print(
            f"\n{mess_prefix}Aligning spike trains of {n_datasets} datasets (w/r to 1st dataset)...{mess_suffix}"
        )
        spike_times = align_timeseries_interpol(spike_times, sync_signals, 30000)
        # Now save the new array
        merged_spike_clusters = npa(zeros=(NspikesTotal), dtype=np.float64)
        merged_spike_times = npa(zeros=(NspikesTotal), dtype=np.uint64)
        cum_Nspikes = 0
        for ds_i in ds_table.index:
            Nspikes = len(spike_times[ds_i])
            merged_spike_clusters[cum_Nspikes : cum_Nspikes + Nspikes] = (
                spike_clusters[ds_i] + 1e-1 * ds_i
            )
            merged_spike_times[cum_Nspikes : cum_Nspikes + Nspikes] = spike_times[ds_i]
            cum_Nspikes += Nspikes
        merged_spike_clusters = merged_spike_clusters[np.argsort(merged_spike_times)]
        merged_spike_times = np.sort(merged_spike_times)
        np.save(dp_merged / (merge_fname_clusters + ".npy"), merged_spike_clusters)
        np.save(dp_merged / (merge_fname_times + ".npy"), merged_spike_times)
        sync_dir = dp_merged / "sync_chan"
        sync_dir.mkdir(exist_ok=True)
        sync_file = (
            sync_dir
            / f"merged_ref_{ds_table.loc[0, 'dataset_name']}.ap_sync_on_samples.npy"
        )
        np.save(sync_file, sync_signals[0])
        print(
            f"\n{mess_prefix}Merged {merge_fname_times} and {merge_fname_clusters} saved at {dp_merged}.{mess_suffix}"
        )

        success_message = "\n--> Merge successful! Use a float u.x in any npyx function to call unit u from dataset x:"
        for ds_i in ds_table.index:
            success_message += (
                f"\n- u.{ds_i} for dataset {ds_table.loc[ds_i,'dataset_name']},"
            )
        success_message = "\033[92;1m" + success_message[:-1] + ".\033[0m"
        print(success_message)

    return dp_merged, ds_table


def ask_syncchan(ons):
    chan_len = "".join([f"chan {k} ({len(v)} events)." for k, v in ons.items()])
    print(f"Data found on sync channels:\n{chan_len}")
    if len(ons) == 1:
        print("Only one sync channel with data -> using it to synchronize probes.\n")
        return list(ons.keys())[0]
    syncchan = None
    while syncchan is None:
        syncchan = input(
            "Which channel shall be used to synchronize probes? (q to restart and pick other channel on previous probe) >>> "
        )
        try:
            syncchan = int(syncchan)
            if syncchan not in ons.keys():
                print(f"!! You need to feed an integer amongst {list(ons.keys())}!")
                syncchan = None
        except:
            if syncchan == "q":
                print("Aborting...")
                return syncchan
            print("!! You need to feed an integer!")
            syncchan = None
        if syncchan is not None:
            if not any(ons[syncchan]):
                print(
                    "!! This sync channel does not have any events! Pick another one!"
                )
                syncchan = None
    return syncchan


def get_ds_table(dp):
    assert assert_multi(dp), "Cannot be ran on a non merged dataset."
    dp = Path(dp)
    ds_table = pd.read_csv(dp / "datasets_table.csv", index_col="dataset_i")
    for dp in ds_table["dp"]:
        assert op.exists(
            dp
        ), f"WARNING you have instanciated this merged dataset from paths of which at least one doesn't exist anymore:{dp}!"
    return ds_table


def get_dataset_id(u):
    """
    Parameters:
        - u: float, of format u.ds_i
    Returns:
        - u: int, unit index
        - ds_i: int, dataset index
    """
    assert assert_float(
        u
    ), "Seems like the unit passed isn't a float - \
        calling this function on a merged dataset is meaningless \
        (cannot tell which dataset the unit belongs to!)."
    return int(round(u % 1 * 10)), int(u)


def assert_same_dataset(U):
    """Asserts if all provided units belong to the same dataset."""
    return all(get_dataset_id(U[0])[0] == get_dataset_id(u)[0] for u in U[1:])


def assert_multi(dp):
    return "merged" in op.basename(dp)


def get_ds_ids(U):
    return (U % 1 * 10).round(0).astype(np.int64)


def get_dataset_ids(dp_pro):
    """
    Parameters:
        - dp_pro: str, path to merged dataset
    Returns:
        - dataset_ids: np array of shape (N_spikes,), indices of dataset of origin for all spikes
    """
    assert assert_multi(dp_pro)
    return get_ds_ids(get_units(dp_pro))


def get_source_dp_u(dp, u):
    """If dp is a merged datapath, returns datapath from source dataset and unit as integer.
    Else, returns dp and u as they are.
    """
    if assert_multi(dp):
        ds_i, u = get_dataset_id(u)
        ds_table = get_ds_table(dp)
        dp = ds_table["dp"][ds_i]

    dp = Path(dp)
    return dp, u
